"""
This block defines a Staff Graded Assignment.  Students are shown a rubric
and invited to upload a file which is then graded by staff.
"""
from __future__ import absolute_import

import json
import logging
import mimetypes
import os
from contextlib import closing
from zipfile import ZipFile

import pkg_resources
import six
import six.moves.urllib.error
import six.moves.urllib.parse
import six.moves.urllib.request

import pytz
from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.core.files import File
from django.core.files.storage import default_storage
from django.template import Context, Template
from django.utils.encoding import force_text
from django.utils.timezone import now as django_now
from django.utils.translation import ugettext as _
from edx_sga.constants import ITEM_TYPE
from edx_sga.showanswer import ShowAnswerXBlockMixin
from edx_sga.tasks import (get_zip_file_name, get_zip_file_path,
                           zip_student_submissions)
from edx_sga.utils import (file_contents_iter, get_file_modified_time_utc,
                           get_file_storage_path, get_sha1,
                           is_finalized_submission, utcnow)
from lms.djangoapps.courseware.models import StudentModule
from safe_lxml import etree
from student.models import user_by_anonymous_id
from submissions import api as submissions_api
from submissions.models import StudentItem as SubmissionsStudent
from submissions.models import Submission
from webob.response import Response
from xblock.core import XBlock
from xblock.exceptions import JsonHandlerError
from xblock.fields import DateTime, Float, Integer, Scope, String
from xblock.fragment import Fragment
from xblockutils.studio_editable import StudioEditableXBlockMixin
from xmodule.contentstore.content import StaticContent
from xmodule.util.duedate import get_extended_due_date

log = logging.getLogger(__name__)


def reify(meth):
    """
    Decorator which caches value so it is only computed once.
    Keyword arguments:
    inst
    """
    def getter(inst):
        """
        Set value to meth name in dict and returns value.
        """
        value = meth(inst)
        inst.__dict__[meth.__name__] = value
        return value
    return property(getter)


class StaffGradedAssignmentXBlock(StudioEditableXBlockMixin, ShowAnswerXBlockMixin, XBlock):
    """
    This block defines a Staff Graded Assignment.  Students are shown a rubric
    and invited to upload a file which is then graded by staff.
    """
    has_score = True
    icon_class = 'problem'
    STUDENT_FILEUPLOAD_MAX_SIZE = 4 * 1000 * 1000  # 4 MB
    editable_fields = ('display_name', 'points', 'weight', 'showanswer', 'solution')

    display_name = String(
        display_name=_("Display Name"),
        default=_('Staff Graded Assignment'),
        scope=Scope.settings,
        help=_("This name appears in the horizontal navigation at the top of "
               "the page.")
    )

    weight = Float(
        display_name=_("Problem Weight"),
        help=_("Defines the number of points each problem is worth. "
               "If the value is not set, the problem is worth the sum of the "
               "option point values."),
        values={"min": 0, "step": .1},
        scope=Scope.settings
    )

    points = Integer(
        display_name=_("Maximum score"),
        help=_("Maximum grade score given to assignment by staff."),
        default=100,
        scope=Scope.settings
    )

    staff_score = Integer(
        display_name=_("Score assigned by non-instructor staff"),
        help=_("Score will need to be approved by instructor before being "
               "published."),
        default=None,
        scope=Scope.settings
    )

    comment = String(
        display_name=_("Instructor comment"),
        default='',
        scope=Scope.user_state,
        help=_("Feedback given to student by instructor.")
    )

    annotated_sha1 = String(
        display_name=_("Annotated SHA1"),
        scope=Scope.user_state,
        default=None,
        help=_("sha1 of the annotated file uploaded by the instructor for "
               "this assignment.")
    )

    annotated_filename = String(
        display_name=_("Annotated file name"),
        scope=Scope.user_state,
        default=None,
        help=_("The name of the annotated file uploaded for this assignment.")
    )

    annotated_mimetype = String(
        display_name=_("Mime type of annotated file"),
        scope=Scope.user_state,
        default=None,
        help=_("The mimetype of the annotated file uploaded for this assignment.")
    )

    annotated_timestamp = DateTime(
        display_name=_("Timestamp"),
        scope=Scope.user_state,
        default=None,
        help=_("When the annotated file was uploaded")
    )

    @classmethod
    def student_upload_max_size(cls):
        """
        returns max file size limit in system
        """
        return getattr(
            settings,
            "STUDENT_FILEUPLOAD_MAX_SIZE",
            cls.STUDENT_FILEUPLOAD_MAX_SIZE
        )

    @classmethod
    def file_size_over_limit(cls, file_obj):
        """
        checks if file size is under limit.
        """
        file_obj.seek(0, os.SEEK_END)
        return file_obj.tell() > cls.student_upload_max_size()

    @classmethod
    def parse_xml(cls, node, runtime, keys, id_generator):
        """
        Override default serialization to handle <solution /> elements
        """
        block = runtime.construct_xblock_from_class(cls, keys)

        for child in node:
            if child.tag == "solution":
                # convert child elements of <solution> into HTML for display
                block.solution = ''.join(etree.tostring(subchild) for subchild in child)

        # Attributes become fields.
        # Note that a solution attribute here will override any solution XML element
        for name, value in node.items():  # lxml has no iteritems
            cls._set_field_if_present(block, name, value, {})

        return block

    def add_xml_to_node(self, node):
        """
        Override default serialization to output solution field as a separate child element.
        """
        super(StaffGradedAssignmentXBlock, self).add_xml_to_node(node)

        if 'solution' in node.attrib:
            # Try outputting it as an XML element if we can
            solution = node.attrib['solution']
            wrapped = "<solution>{}</solution>".format(solution)
            try:
                child = etree.fromstring(wrapped)
            except:  # pylint: disable=bare-except
                # Parsing exception, leave the solution as an attribute
                pass
            else:
                node.append(child)
                del node.attrib['solution']

    @XBlock.json_handler
    def save_sga(self, data, suffix=''):
        # pylint: disable=unused-argument
        """
        Persist block data when updating settings in studio.
        """
        self.display_name = data.get('display_name', self.display_name)

        # Validate points before saving
        points = data.get('points', self.points)
        # Check that we are an int
        try:
            points = int(points)
        except ValueError:
            raise JsonHandlerError(400, 'Points must be an integer')
        # Check that we are positive
        if points < 0:
            raise JsonHandlerError(400, 'Points must be a positive integer')
        self.points = points

        # Validate weight before saving
        weight = data.get('weight', self.weight)
        # Check that weight is a float.
        if weight:
            try:
                weight = float(weight)
            except ValueError:
                raise JsonHandlerError(400, 'Weight must be a decimal number')
            # Check that we are positive
            if weight < 0:
                raise JsonHandlerError(
                    400, 'Weight must be a positive decimal number'
                )
        self.weight = weight

    @XBlock.handler
    def upload_assignment(self, request, suffix=''):
        # pylint: disable=unused-argument, protected-access
        """
        Save a students submission file.
        """
        require(self.upload_allowed())
        user = self.get_real_user()
        require(user)
        upload = request.params['assignment']
        sha1 = get_sha1(upload.file)
        if self.file_size_over_limit(upload.file):
            raise JsonHandlerError(
                413, 'Unable to upload file. Max size limit is {size}'.format(
                    size=self.student_upload_max_size()
                )
            )
        # Uploading an assignment represents a change of state with this user in this block,
        # so we need to ensure that the user has a StudentModule record, which represents that state.
        self.get_or_create_student_module(user)
        answer = {
            "sha1": sha1,
            "filename": upload.file.name,
            "mimetype": mimetypes.guess_type(upload.file.name)[0],
            "finalized": False
        }
        student_item_dict = self.get_student_item_dict()
        submissions_api.create_submission(student_item_dict, answer)
        path = self.file_storage_path(sha1, upload.file.name)
        log.info("Saving file: %s at path: %s for user: %s", upload.file.name, path, user.username)
        if default_storage.exists(path):
            # save latest submission
            default_storage.delete(path)
        default_storage.save(path, File(upload.file))
        return Response(json_body=self.student_state())

    @XBlock.handler
    def finalize_uploaded_assignment(self, request, suffix=''):
        # pylint: disable=unused-argument
        """
        Finalize a student's uploaded submission. This prevents further uploads for the
        given block, and makes the submission available to instructors for grading
        """
        submission_data = self.get_submission()
        require(self.upload_allowed(submission_data=submission_data))
        # Editing the Submission record directly since the API doesn't support it
        submission = Submission.objects.get(uuid=submission_data['uuid'])
        if not submission.answer.get('finalized'):
            submission.answer['finalized'] = True
            submission.submitted_at = django_now()
            submission.save()
        return Response(json_body=self.student_state())

    @XBlock.handler
    def staff_upload_annotated(self, request, suffix=''):
        # pylint: disable=unused-argument
        """
        Save annotated assignment from staff.
        """
        require(self.is_course_staff())
        upload = request.params['annotated']
        sha1 = get_sha1(upload.file)
        if self.file_size_over_limit(upload.file):
            raise JsonHandlerError(
                413, 'Unable to upload file. Max size limit is {size}'.format(
                    size=self.student_upload_max_size()
                )
            )
        module = self.get_student_module(request.params['module_id'])
        state = json.loads(module.state)
        state['annotated_sha1'] = sha1
        state['annotated_filename'] = filename = upload.file.name
        state['annotated_mimetype'] = mimetypes.guess_type(upload.file.name)[0]
        state['annotated_timestamp'] = utcnow().strftime(
            DateTime.DATETIME_FORMAT
        )
        path = self.file_storage_path(sha1, filename)
        if not default_storage.exists(path):
            default_storage.save(path, File(upload.file))
        module.state = json.dumps(state)
        module.save()
        log.info(
            "staff_upload_annotated for course:%s module:%s student:%s ",
            module.course_id,
            module.module_state_key,
            module.student.username
        )
        return Response(json_body=self.staff_grading_data())

    @XBlock.handler
    def download_assignment(self, request, suffix=''):
        # pylint: disable=unused-argument
        """
        Fetch student assignment from storage and return it.
        """
        answer = self.get_submission()['answer']
        path = self.file_storage_path(answer['sha1'], answer['filename'])
        return self.download(path, answer['mimetype'], answer['filename'])

    @XBlock.handler
    def download_annotated(self, request, suffix=''):
        # pylint: disable=unused-argument
        """
        Fetch assignment with staff annotations from storage and return it.
        """
        path = self.file_storage_path(
            self.annotated_sha1,
            self.annotated_filename,
        )
        return self.download(
            path,
            self.annotated_mimetype,
            self.annotated_filename
        )

    @XBlock.handler
    def staff_download(self, request, suffix=''):
        # pylint: disable=unused-argument
        """
        Return an assignment file requested by staff.
        """
        require(self.is_course_staff())
        submission = self.get_submission(request.params['student_id'])
        answer = submission['answer']
        path = self.file_storage_path(answer['sha1'], answer['filename'])
        return self.download(
            path,
            answer['mimetype'],
            answer['filename'],
            require_staff=True
        )

    @XBlock.handler
    def staff_download_annotated(self, request, suffix=''):
        # pylint: disable=unused-argument
        """
        Return annotated assignment file requested by staff.
        """
        require(self.is_course_staff())
        module = self.get_student_module(request.params['module_id'])
        state = json.loads(module.state)
        path = self.file_storage_path(
            state['annotated_sha1'],
            state['annotated_filename']
        )
        return self.download(
            path,
            state['annotated_mimetype'],
            state['annotated_filename'],
            require_staff=True
        )

    @XBlock.handler
    def get_staff_grading_data(self, request, suffix=''):
        # pylint: disable=unused-argument
        """
        Return the html for the staff grading view
        """
        require(self.is_course_staff())
        return Response(json_body=self.staff_grading_data())

    @XBlock.handler
    def enter_grade(self, request, suffix=''):
        # pylint: disable=unused-argument
        """
        Persist a score for a student given by staff.
        """
        require(self.is_course_staff())
        score = request.params.get('grade', None)
        module = self.get_student_module(request.params['module_id'])
        if not score:
            return Response(
                json_body=self.validate_score_message(
                    module.course_id,
                    module.student.username
                )
            )

        state = json.loads(module.state)
        try:
            score = int(score)
        except ValueError:
            return Response(
                json_body=self.validate_score_message(
                    module.course_id,
                    module.student.username
                )
            )

        if self.is_instructor():
            uuid = request.params['submission_id']
            submissions_api.set_score(uuid, score, self.max_score())
        else:
            state['staff_score'] = score
        state['comment'] = request.params.get('comment', '')
        module.state = json.dumps(state)
        module.save()
        log.info(
            "enter_grade for course:%s module:%s student:%s",
            module.course_id,
            module.module_state_key,
            module.student.username
        )

        return Response(json_body=self.staff_grading_data())

    @XBlock.handler
    def remove_grade(self, request, suffix=''):
        # pylint: disable=unused-argument
        """
        Reset a students score request by staff.
        """
        require(self.is_course_staff())
        student_id = request.params['student_id']
        submissions_api.reset_score(
            student_id,
            self.block_course_id,
            self.block_id
        )
        module = self.get_student_module(request.params['module_id'])
        state = json.loads(module.state)
        state['staff_score'] = None
        state['comment'] = ''
        state['annotated_sha1'] = None
        state['annotated_filename'] = None
        state['annotated_mimetype'] = None
        state['annotated_timestamp'] = None
        module.state = json.dumps(state)
        module.save()
        log.info(
            "remove_grade for course:%s module:%s student:%s",
            module.course_id,
            module.module_state_key,
            module.student.username
        )
        return Response(json_body=self.staff_grading_data())

    @XBlock.handler
    def prepare_download_submissions(self, request, suffix=''):  # pylint: disable=unused-argument
        """
        Runs a async task that collects submissions in background and zip them.
        """
        # pylint: disable=no-member
        require(self.is_course_staff())
        user = self.get_real_user()
        require(user)
        zip_file_ready = False
        location = six.text_type(self.location)

        if self.is_zip_file_available(user):
            log.info("Zip file already available for block: %s for instructor: %s", location, user.username)
            assignments = self.get_sorted_submissions()
            if assignments:
                last_assignment_date = assignments[0]['timestamp'].astimezone(pytz.utc)
                zip_file_path = get_zip_file_path(
                    user.username,
                    self.block_course_id,
                    self.block_id,
                    self.location
                )
                zip_file_time = get_file_modified_time_utc(zip_file_path)
                log.info(
                    "Zip file modified time: %s, last zip file time: %s for block: %s for instructor: %s",
                    last_assignment_date,
                    zip_file_time,
                    location,
                    user.username
                )
                # if last zip file is older the last submission then recreate task
                if zip_file_time >= last_assignment_date:
                    zip_file_ready = True

                # check if some one reset submission. If yes the recreate zip file
                assignment_count = len(assignments)
                if self.count_archive_files(user) != assignment_count:
                    zip_file_ready = False

        if not zip_file_ready:
            log.info("Creating new zip file for block: %s for instructor: %s", location, user.username)
            zip_student_submissions.delay(
                self.block_course_id,
                self.block_id,
                location,
                user.username
            )

        return Response(json_body={
            "downloadable": zip_file_ready
        })

    @XBlock.handler
    def download_submissions(self, request, suffix=''):  # pylint: disable=unused-argument
        """
        Api for downloading zip file which consist of all students submissions.
        """
        # pylint: disable=no-member
        require(self.is_course_staff())
        user = self.get_real_user()
        require(user)
        try:
            zip_file_path = get_zip_file_path(
                user.username,
                self.block_course_id,
                self.block_id,
                self.location
            )
            zip_file_name = get_zip_file_name(
                user.username,
                self.block_course_id,
                self.block_id
            )
            return Response(
                app_iter=file_contents_iter(zip_file_path),
                content_type='application/zip',
                content_disposition="attachment; filename=" + zip_file_name
            )
        except IOError:
            return Response(
                "Sorry, submissions cannot be found. Press Collect ALL Submissions button or"
                " contact {} if you issue is consistent".format(settings.TECH_SUPPORT_EMAIL),
                status_code=404
            )

    @XBlock.handler
    def download_submissions_status(self, request, suffix=''):  # pylint: disable=unused-argument
        """
        returns True if zip file is available for download
        """
        require(self.is_course_staff())
        user = self.get_real_user()
        require(user)
        return Response(
            json_body={
                "zip_available": self.is_zip_file_available(user)
            }
        )

    def student_view(self, context=None):
        # pylint: disable=no-member
        """
        The primary view of the StaffGradedAssignmentXBlock, shown to students
        when viewing courses.
        """
        context = {
            "student_state": json.dumps(self.student_state()),
            "id": self.location.name.replace('.', '_'),
            "max_file_size": self.student_upload_max_size(),
            "support_email": settings.TECH_SUPPORT_EMAIL
        }
        if self.show_staff_grading_interface():
            context['is_course_staff'] = True
            self.update_staff_debug_context(context)

        fragment = Fragment()
        fragment.add_content(
            render_template(
                'templates/staff_graded_assignment/show.html',
                context
            )
        )
        fragment.add_css(_resource("static/css/edx_sga.css"))
        fragment.add_javascript(_resource("static/js/src/edx_sga.js"))
        fragment.add_javascript(_resource("static/js/src/jquery.tablesorter.min.js"))
        fragment.initialize_js('StaffGradedAssignmentXBlock')
        return fragment

    def studio_view(self, context=None):  # pylint: disable=useless-super-delegation
        """
        Render a form for editing this XBlock
        """
        # this method only exists to provide context=None for backwards compat
        return super(StaffGradedAssignmentXBlock, self).studio_view(context)

    def clear_student_state(self, *args, **kwargs):
        # pylint: disable=unused-argument
        """
        For a given user, clears submissions and uploaded files for this XBlock.

        Staff users are able to delete a learner's state for a block in LMS. When that capability is
        used, the block's "clear_student_state" function is called if it exists.
        """
        student_id = kwargs['user_id']
        for submission in submissions_api.get_submissions(
                self.get_student_item_dict(student_id)
        ):
            submission_file_sha1 = submission['answer'].get('sha1')
            submission_filename = submission['answer'].get('filename')
            submission_file_path = self.file_storage_path(submission_file_sha1, submission_filename)
            if default_storage.exists(submission_file_path):
                default_storage.delete(submission_file_path)
            submissions_api.reset_score(
                student_id,
                self.block_course_id,
                self.block_id,
                clear_state=True
            )

    def max_score(self):
        """
        Return the maximum score possible.
        """
        return self.points

    @reify
    def block_id(self):
        """
        Return the usage_id of the block.
        """
        return six.text_type(self.scope_ids.usage_id)

    @reify
    def block_course_id(self):
        """
        Return the course_id of the block.
        """
        return six.text_type(self.course_id)

    def get_student_item_dict(self, student_id=None):
        # pylint: disable=no-member
        """
        Returns dict required by the submissions app for creating and
        retrieving submissions for a particular student.
        """
        if student_id is None:
            student_id = self.xmodule_runtime.anonymous_student_id
            assert student_id != (
                'MOCK', "Forgot to call 'personalize' in test."
            )
        return {
            "student_id": student_id,
            "course_id": self.block_course_id,
            "item_id": self.block_id,
            "item_type": ITEM_TYPE,
        }

    def get_submission(self, student_id=None):
        """
        Get student's most recent submission.
        """
        submissions = submissions_api.get_submissions(
            self.get_student_item_dict(student_id)
        )
        if submissions:
            # If I understand docs correctly, most recent submission should
            # be first
            return submissions[0]

    def get_score(self, student_id=None):
        """
        Return student's current score.
        """
        score = submissions_api.get_score(
            self.get_student_item_dict(student_id)
        )
        if score:
            return score['points_earned']

    @reify
    def score(self):
        """
        Return score from submissions.
        """
        return self.get_score()

    def update_staff_debug_context(self, context):
        # pylint: disable=no-member
        """
        Add context info for the Staff Debug interface.
        """
        published = self.start
        context['is_released'] = published and published < utcnow()
        context['location'] = self.location
        context['category'] = type(self).__name__
        context['fields'] = [
            (name, field.read_from(self))
            for name, field in self.fields.items()]

    def get_student_module(self, module_id):
        """
        Returns a StudentModule that matches the given id

        Args:
            module_id (int): The module id

        Returns:
            StudentModule: A StudentModule object
        """
        return StudentModule.objects.get(pk=module_id)

    def get_or_create_student_module(self, user):
        """
        Gets or creates a StudentModule for the given user for this block

        Returns:
            StudentModule: A StudentModule object
        """
        # pylint: disable=no-member
        student_module, created = StudentModule.objects.get_or_create(
            course_id=self.course_id,
            module_state_key=self.location,
            student=user,
            defaults={
                'state': '{}',
                'module_type': self.category,
            }
        )
        if created:
            log.info(
                "Created student module %s [course: %s] [student: %s]",
                student_module.module_state_key,
                student_module.course_id,
                student_module.student.username
            )
        return student_module

    def student_state(self):
        """
        Returns a JSON serializable representation of student's state for
        rendering in client view.
        """
        submission = self.get_submission()
        if submission:
            uploaded = {"filename": submission['answer']['filename']}
        else:
            uploaded = None

        if self.annotated_sha1:
            annotated = {"filename": force_text(self.annotated_filename)}
        else:
            annotated = None

        score = self.score
        if score is not None:
            graded = {'score': score, 'comment': force_text(self.comment)}
        else:
            graded = None

        if self.answer_available():
            solution = self.runtime.replace_urls(force_text(self.solution))
        else:
            solution = ''
        # pylint: disable=no-member
        return {
            "display_name": force_text(self.display_name),
            "uploaded": uploaded,
            "annotated": annotated,
            "graded": graded,
            "max_score": self.max_score(),
            "upload_allowed": self.upload_allowed(submission_data=submission),
            "solution": solution,
            "base_asset_url": StaticContent.get_base_url_path_for_course_assets(self.location.course_key),
        }

    def staff_grading_data(self):
        """
        Return student assignment information for display on the
        grading screen.
        """
        def get_student_data():
            # pylint: disable=no-member
            """
            Returns a dict of student assignment information along with
            annotated file name, student id and module id, this
            information will be used on grading screen
            """
            # Submissions doesn't have API for this, just use model directly.
            students = SubmissionsStudent.objects.filter(
                course_id=self.course_id,
                item_id=self.block_id)
            for student in students:
                submission = self.get_submission(student.student_id)
                if not submission:
                    continue
                user = user_by_anonymous_id(student.student_id)
                student_module = self.get_or_create_student_module(user)
                state = json.loads(student_module.state)
                score = self.get_score(student.student_id)
                approved = score is not None
                if score is None:
                    score = state.get('staff_score')
                    needs_approval = score is not None
                else:
                    needs_approval = False
                instructor = self.is_instructor()
                yield {
                    'module_id': student_module.id,
                    'student_id': student.student_id,
                    'submission_id': submission['uuid'],
                    'username': student_module.student.username,
                    'fullname': student_module.student.profile.name,
                    'filename': submission['answer']["filename"],
                    'timestamp': submission['created_at'].strftime(
                        DateTime.DATETIME_FORMAT
                    ),
                    'score': score,
                    'approved': approved,
                    'needs_approval': instructor and needs_approval,
                    'may_grade': instructor or not approved,
                    'annotated': force_text(state.get("annotated_filename", '')),
                    'comment': force_text(state.get("comment", '')),
                    'finalized': is_finalized_submission(submission_data=submission)
                }

        return {
            'assignments': list(get_student_data()),
            'max_score': self.max_score(),
            'display_name': force_text(self.display_name)
        }

    def get_sorted_submissions(self):
        """returns student recent assignments sorted on date"""
        assignments = []
        submissions = submissions_api.get_all_submissions(
            self.course_id,
            self.block_id,
            ITEM_TYPE
        )

        for submission in submissions:
            if is_finalized_submission(submission_data=submission):
                assignments.append({
                    'submission_id': submission['uuid'],
                    'filename': submission['answer']["filename"],
                    'timestamp': submission['submitted_at'] or submission['created_at']
                })

        assignments.sort(
            key=lambda assignment: assignment['timestamp'], reverse=True
        )
        return assignments

    def download(self, path, mime_type, filename, require_staff=False):
        """
        Return a file from storage and return in a Response.
        """
        try:
            content_disposition = "attachment; filename*=UTF-8''"
            content_disposition += six.moves.urllib.parse.quote(filename.encode('utf-8'))
            output = Response(
                app_iter=file_contents_iter(path),
                content_type=mime_type,
                content_disposition=content_disposition
            )
            return output
        except IOError:
            if require_staff:
                return Response(
                    "Sorry, assignment {} cannot be found at"
                    " {}. Please contact {}".format(
                        filename.encode('utf-8'), path, settings.TECH_SUPPORT_EMAIL
                    ),
                    status_code=404
                )
            return Response(
                "Sorry, the file you uploaded, {}, cannot be"
                " found. Please try uploading it again or contact"
                " course staff".format(filename.encode('utf-8')),
                status_code=404
            )

    def validate_score_message(self, course_id, username):  # lint-amnesty, pylint: disable=missing-docstring
        # pylint: disable=no-member
        log.error(
            "enter_grade: invalid grade submitted for course:%s module:%s student:%s",
            course_id,
            self.location,
            username
        )
        return {
            "error": "Please enter valid grade"
        }

    def is_course_staff(self):
        # pylint: disable=no-member
        """
         Check if user is course staff.
        """
        return getattr(self.xmodule_runtime, 'user_is_staff', False)

    def is_instructor(self):
        # pylint: disable=no-member
        """
        Check if user role is instructor.
        """
        return self.xmodule_runtime.get_user_role() == 'instructor'

    def show_staff_grading_interface(self):
        """
        Return if current user is staff and not in studio.
        """
        in_studio_preview = self.scope_ids.user_id is None
        return self.is_course_staff() and not in_studio_preview

    def past_due(self):
        """
        Return whether due date has passed.
        """
        due = get_extended_due_date(self)
        try:
            graceperiod = self.graceperiod
        except AttributeError:
            # graceperiod and due are defined in InheritanceMixin
            # It's used automatically in edX but the unit tests will need to mock it out
            graceperiod = None

        if graceperiod is not None and due:
            close_date = due + graceperiod
        else:
            close_date = due

        if close_date is not None:
            return utcnow() > close_date
        return False

    def upload_allowed(self, submission_data=None):
        """
        Return whether student is allowed to upload an assignment.
        """
        submission_data = submission_data if submission_data is not None else self.get_submission()
        return (
            not self.past_due() and
            self.score is None and
            not is_finalized_submission(submission_data)
        )

    def file_storage_path(self, file_hash, original_filename):
        # pylint: disable=no-member
        """
        Helper method to get the path of an uploaded file
        """
        return get_file_storage_path(self.location, file_hash, original_filename)

    def is_zip_file_available(self, user):
        """
        returns True if zip file exists.
        """
        # pylint: disable=no-member
        zip_file_path = get_zip_file_path(
            user.username,
            self.block_course_id,
            self.block_id,
            self.location
        )
        return True if default_storage.exists(zip_file_path) else False

    def count_archive_files(self, user):
        """
        returns number of files archive in zip.
        """
        # pylint: disable=no-member
        zip_file_path = get_zip_file_path(
            user.username,
            self.block_course_id,
            self.block_id,
            self.location
        )
        with default_storage.open(zip_file_path, 'rb') as zip_file:
            with closing(ZipFile(zip_file)) as archive:
                return len(archive.infolist())

    def get_real_user(self):
        """returns session user"""
        # pylint: disable=no-member
        return self.runtime.get_real_user(self.xmodule_runtime.anonymous_student_id)

    def correctness_available(self):
        """
        For SGA is_correct just means the user submitted the problem, which we always know one way or the other
        """
        return True

    def is_past_due(self):
        """
        Is it now past this problem's due date?
        """
        return self.past_due()

    def is_correct(self):
        """
        For SGA we show the answer as soon as we know the user has given us their submission
        """
        return self.has_attempted()

    def has_attempted(self):
        """
        True if the student has already attempted this problem
        """
        submission = self.get_submission()
        if not submission:
            return False
        return submission['answer']['finalized']

    def can_attempt(self):
        """
        True if the student can attempt the problem
        """
        return not self.has_attempted()

    def runtime_user_is_staff(self):
        """
        Is the logged in user a staff user?
        """
        return self.is_course_staff()


def _resource(path):  # pragma: NO COVER
    """
    Handy helper for getting resources from our kit.
    """
    data = pkg_resources.resource_string(__name__, path)
    return data.decode("utf8")


def load_resource(resource_path):  # pragma: NO COVER
    """
    Gets the content of a resource
    """
    resource_content = pkg_resources.resource_string(__name__, resource_path)
    return six.text_type(resource_content.decode("utf8"))


def render_template(template_path, context=None):  # pragma: NO COVER
    """
    Evaluate a template by resource path, applying the provided context.
    """
    if context is None:
        context = {}

    template_str = load_resource(template_path)
    template = Template(template_str)
    return template.render(Context(context))


def require(assertion):
    """
    Raises PermissionDenied if assertion is not true.
    """
    if not assertion:
        raise PermissionDenied



"""
Utility functions for the SGA XBlock
"""
from __future__ import absolute_import

import datetime
import hashlib
import os
import time
from functools import partial

import six

import pytz
from django.conf import settings
from django.core.files.storage import default_storage
from edx_sga.constants import BLOCK_SIZE


def utcnow():
    """
    Get current date and time in UTC
    """
    return datetime.datetime.now(tz=pytz.utc)


def is_finalized_submission(submission_data):
    """
    Helper function to determine whether or not a Submission was finalized by the student
    """
    if submission_data and submission_data.get('answer') is not None:
        return submission_data['answer'].get('finalized', True)
    return False


def get_file_modified_time_utc(file_path):
    """
    Gets the UTC timezone-aware modified time of a file at the given file path
    """
    file_timezone = (
        # time.tzname returns a 2 element tuple:
        #   (local non-DST timezone, e.g.: 'EST', local DST timezone, e.g.: 'EDT')
        pytz.timezone(time.tzname[0])
        if settings.DEFAULT_FILE_STORAGE == 'django.core.files.storage.FileSystemStorage'
        else pytz.utc
    )
    return file_timezone.localize(
        default_storage.modified_time(file_path)
    ).astimezone(
        pytz.utc
    )


def get_sha1(file_descriptor):
    """
    Get file hex digest (fingerprint).
    """
    sha1 = hashlib.sha1()
    for block in iter(partial(file_descriptor.read, BLOCK_SIZE), b''):
        sha1.update(block)
    file_descriptor.seek(0)
    return sha1.hexdigest()


def get_file_storage_path(locator, file_hash, original_filename):
    """
    Returns the file path for an uploaded SGA submission file
    """
    return (
        six.u(
            '{loc.org}/{loc.course}/{loc.block_type}/{loc.block_id}/{file_hash}{ext}'
        ).format(
            loc=locator,
            file_hash=file_hash,
            ext=os.path.splitext(original_filename)[1]
        )
    )


def file_contents_iter(file_path):
    """
    Returns an iterator over the contents of a file located at the given file path
    """
    file_descriptor = default_storage.open(file_path)
    return iter(partial(file_descriptor.read, BLOCK_SIZE), b'')



"""
Django settings for edx_sga project.
"""

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
from __future__ import absolute_import

import os

DEFAULT_FILE_STORAGE = 'django.core.files.storage.FileSystemStorage'
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
SECRET_KEY = 'k(g$9p*x^a=t_lf$%=b*3$$ipg1g0tm!^pws0@g)u+v@&$657n'
DEBUG = True

ALLOWED_HOSTS = []
INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'workbench',
)
MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.auth.middleware.SessionAuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'django.middleware.security.SecurityMiddleware',
)
TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]
WSGI_APPLICATION = 'edx_sga.wsgi.application'


# Database
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}

# Internationalization
LANGUAGE_CODE = 'en-us'
TIME_ZONE = 'UTC'
USE_I18N = True
USE_L10N = True
USE_TZ = True
STATIC_URL = '/static/'
TECH_SUPPORT_EMAIL = 'support@example.com'
# For xblock-sdk workbench
WORKBENCH = {}



# -*- coding: utf-8 -*-
"""
Tests for SGA utility functions
"""
from __future__ import absolute_import

import pytest

import pytz
from edx_sga.tests.common import is_near_now
from edx_sga.utils import is_finalized_submission, utcnow


@pytest.mark.parametrize(
    'submission_data,expected_value', [
        ({'answer': {'finalized': True}}, True),
        ({'answer': {'filename': 'file.txt'}}, True),
        ({'answer': {}}, True),
        ({'answer': {'finalized': False}}, False),
        ({'answer': None}, False),
        ({}, False),
    ]
)
def test_is_finalized_submission(submission_data, expected_value):
    """Test for is_finalized_submission"""
    assert is_finalized_submission(submission_data) is expected_value


def test_utcnow():
    """
    tznow should return a datetime object in UTC
    """
    now = utcnow()
    assert is_near_now(now)
    assert now.tzinfo.zone == pytz.utc.zone



# -*- coding: utf-8 -*-
"""
Tests for SGA
"""
from __future__ import absolute_import

import datetime
import json
import mimetypes
import os
import uuid

import mock
import pytest
import six
from six.moves import range

import pytz
from ddt import data, ddt, unpack
from django.conf import settings
from django.contrib.auth.models import User
from django.utils.timezone import now as django_now
from edx_sga.tests.common import DummyResource, TempfileMixin
from opaque_keys.edx.locations import Location
from opaque_keys.edx.locator import CourseLocator
from workbench.runtime import WorkbenchRuntime
from xblock.field_data import DictFieldData
from xblock.fields import DateTime

try:
    # Python 2
    import six.moves.builtins as builtins  # pylint: disable=ungrouped-imports
except ImportError:
    # Python 3
    import builtins  # pylint: disable=ungrouped-imports


SHA1 = 'da39a3ee5e6b4b0d3255bfef95601890afd80709'
UUID = '8c4b765745f746f7a128470842211601'


pytestmark = pytest.mark.django_db  # pylint: disable=invalid-name


def fake_get_submission(**kwargs):
    """returns fake submission data"""
    answer = {
        "sha1": SHA1,
        "filename": kwargs.get("filename", "file.txt"),
        "mimetype": kwargs.get("mimetype", "mime/type"),
    }
    if kwargs.get("finalized"):
        answer["finalized"] = kwargs.get("finalized")
    return {
        "answer": answer,
        "uuid": UUID,
        "submitted_at": kwargs.get("submitted_at", None)
    }


def fake_upload_submission(upload):
    """returns fake submission data with values calculated from an upload object"""
    return fake_get_submission(
        filename=upload.file.name,
        mimetype=mimetypes.guess_type(upload.file.name)[0]
    )


def fake_student_module():
    """dummy representation of xblock class"""
    return mock.Mock(
        course_id=CourseLocator(org='foo', course='baz', run='bar'),
        module_state_key="foo",
        student=mock.Mock(username="fred6", is_staff=False, password="test"),
        state='{"display_name": "Staff Graded Assignment"}',
        save=mock.Mock()
    )


class FakeWorkbenchRuntime(WorkbenchRuntime):
    """Override for testing purposes"""

    anonymous_student_id = 'MOCK'
    user_is_staff = True

    def __init__(self, *args, **kwargs):
        super(FakeWorkbenchRuntime, self).__init__(*args, **kwargs)

        User.objects.create(username=self.anonymous_student_id)

    def get_real_user(self, username):
        """Get the real user"""
        return User.objects.get(username=username)


@ddt
class StaffGradedAssignmentMockedTests(TempfileMixin):
    """
    Create a SGA block with mock data.
    """
    @classmethod
    def setUpClass(cls):
        super(StaffGradedAssignmentMockedTests, cls).setUpClass()
        cls.set_up_temp_directory()

    @classmethod
    def tearDownClass(cls):
        super(StaffGradedAssignmentMockedTests, cls).tearDownClass()
        cls.tear_down_temp_directory()

    def setUp(self):
        """
        Creates a test course ID, mocks the runtime, and creates a fake storage
        engine for use in all tests
        """
        super(StaffGradedAssignmentMockedTests, self).setUp()

        # fakes imports
        real_import = builtins.__import__

        def fake_import(name, *args, **kwargs):
            """mock imported object if not it is not available"""
            try:
                return real_import(name, *args, **kwargs)
            except ImportError:
                for module in ('courseware', 'lms', 'student', 'xmodule'):
                    if name.startswith("{}.".format(module)) or name == module:
                        return mock.Mock()
                if name == 'safe_lxml':
                    return real_import('lxml', *args, **kwargs)
                raise
        builtins.__import__ = fake_import

        def restore_import():
            """restore builtin importer"""
            builtins.__import__ = real_import

        self.addCleanup(restore_import)

        self.course_id = CourseLocator(org='foo', course='baz', run='bar')
        self.runtime = FakeWorkbenchRuntime()
        self.scope_ids = mock.Mock()
        self.staff = mock.Mock(return_value={
            "password": "test",
            "username": "tester",
            "is_staff": True
        })

    def make_xblock(self, display_name=None, **kwargs):
        """
        Creates a XBlock SGA for testing purpose.
        """
        from edx_sga.sga import StaffGradedAssignmentXBlock as cls
        field_data = DictFieldData(kwargs)
        block = cls(self.runtime, field_data, self.scope_ids)
        block.location = Location(
            'foo', 'bar', 'baz', 'category', 'name', 'revision'
        )

        block.xmodule_runtime = self.runtime
        block.course_id = self.course_id
        block.scope_ids.usage_id = "i4x://foo/bar/category/name"
        block.category = 'problem'

        if display_name:
            block.display_name = display_name

        block.start = datetime.datetime(2010, 5, 12, 2, 42, tzinfo=pytz.utc)
        return block

    def test_ctor(self):
        """
        Test points are set correctly.
        """
        block = self.make_xblock(points=10)
        assert block.display_name == "Staff Graded Assignment"
        assert block.points == 10

    def test_max_score(self):
        """
        Text max score is set correctly.
        """
        block = self.make_xblock(points=20)
        assert block.max_score() == 20

    def test_max_score_integer(self):
        """
        Test assigning a float max score is rounded to nearest integer.
        """
        block = self.make_xblock(points=20.4)
        assert block.max_score() == 20

    def personalize_upload(self, block, upload):
        """
        Set values on block from file upload.
        """
        now = datetime.datetime.utcnow().replace(tzinfo=pytz.timezone(getattr(settings, "TIME_ZONE", pytz.utc.zone)))
        block.annotated_mimetype = mimetypes.guess_type(upload.file.name)[0]
        block.annotated_filename = upload.file.name.encode('utf-8')
        block.annotated_sha1 = SHA1
        block.annotated_timestamp = now.strftime(
            DateTime.DATETIME_FORMAT
        )

    @mock.patch('edx_sga.sga._resource', DummyResource)
    @mock.patch('edx_sga.sga.render_template')
    @mock.patch('edx_sga.sga.Fragment')
    def test_student_view(self, fragment, render_template):
        # pylint: disable=unused-argument
        """
        Test student view renders correctly.
        """
        block = self.make_xblock("Custom name")

        with mock.patch(
            'edx_sga.sga.StaffGradedAssignmentXBlock.get_submission',
            return_value={}
        ), mock.patch(
            'edx_sga.sga.StaffGradedAssignmentXBlock.student_state',
            return_value={
                'uploaded': None,
                'annotated': None,
                'upload_allowed': True,
                'max_score': 100,
                'graded': None
            }
        ):
            fragment = block.student_view()
            assert render_template.called is True
            template_arg = render_template.call_args[0][0]
            assert template_arg == 'templates/staff_graded_assignment/show.html'
            context = render_template.call_args[0][1]
            assert context['is_course_staff'] is True
            assert context['id'] == 'name'
            student_state = json.loads(context['student_state'])
            assert student_state['uploaded'] is None
            assert student_state['annotated'] is None
            assert student_state['upload_allowed'] is True
            assert student_state['max_score'] == 100
            assert student_state['graded'] is None
            fragment.add_css.assert_called_once_with(
                DummyResource("static/css/edx_sga.css"))
            fragment.initialize_js.assert_called_once_with(
                "StaffGradedAssignmentXBlock")

    @mock.patch('edx_sga.sga._resource', DummyResource)
    @mock.patch('edx_sga.sga.StaffGradedAssignmentXBlock.upload_allowed')
    @mock.patch('edx_sga.sga.StaffGradedAssignmentXBlock.get_score')
    @mock.patch('edx_sga.sga.render_template')
    @mock.patch('edx_sga.sga.Fragment')
    def test_student_view_with_score(self, fragment, render_template, get_score, upload_allowed):
        # pylint: disable=unused-argument
        """
        Tests scores are displayed correctly on student view.
        """
        block = self.make_xblock()
        get_score.return_value = 10
        upload_allowed.return_value = True
        block.comment = "ok"

        with self.dummy_upload('foo.txt') as (upload, _):
            with mock.patch(
                'submissions.api.create_submission',
            ) as mocked_create_submission, mock.patch(
                'edx_sga.sga.StaffGradedAssignmentXBlock.student_state', return_value={}
            ), mock.patch(
                'edx_sga.sga.StaffGradedAssignmentXBlock.get_or_create_student_module',
                return_value=fake_student_module()
            ):
                block.upload_assignment(mock.Mock(params={'assignment': upload}))
            assert mocked_create_submission.called is True

            with mock.patch(
                'edx_sga.sga.StaffGradedAssignmentXBlock.get_submission',
                return_value=fake_upload_submission(upload)
            ), mock.patch(
                'edx_sga.sga.StaffGradedAssignmentXBlock.student_state',
                return_value={
                    'graded': {'comment': 'ok', 'score': 10},
                    'uploaded': {'filename': 'foo.txt'},
                    'max_score': 100
                }
            ):
                fragment = block.student_view()
                assert render_template.called is True
                template_arg = render_template.call_args[0][0]
                assert template_arg == 'templates/staff_graded_assignment/show.html'
                context = render_template.call_args[0][1]
                assert context['is_course_staff'] is True
                assert context['id'] == 'name'
                student_state = json.loads(context['student_state'])
                assert student_state['uploaded'] == {'filename': 'foo.txt'}
                assert student_state['graded'] == {'comment': 'ok', 'score': 10}
                assert student_state['max_score'] == 100

                fragment.add_css.assert_called_once_with(
                    DummyResource("static/css/edx_sga.css"))
                fragment.initialize_js.assert_called_once_with(
                    "StaffGradedAssignmentXBlock")

    def test_studio_view(self):
        """
        Test studio view is using the StudioEditableXBlockMixin function
        """
        with mock.patch('edx_sga.sga.StudioEditableXBlockMixin.studio_view') as studio_view_mock:
            block = self.make_xblock()
            block.studio_view()
        studio_view_mock.assert_called_once_with(None)

    def test_save_sga(self):
        """
        Tests save SGA  block on studio
        """
        def weights_positive_float_test():
            """
            tests weight is non negative float.
            """
            orig_weight = 11.0
            # Test negative weight doesn't work
            block.save_sga(mock.Mock(method="POST", body=json.dumps({
                "display_name": "Test Block",
                "points": '100',
                "weight": -10.0}).encode('utf-8')))
            assert block.weight == orig_weight

            # Test string weight doesn't work
            block.save_sga(mock.Mock(method="POST", body=json.dumps({
                "display_name": "Test Block",
                "points": '100',
                "weight": "a"}).encode('utf-8')))
            assert block.weight == orig_weight

        def point_positive_int_test():
            """
            Tests point is positive number.
            """
            # Test negative doesn't work
            block.save_sga(mock.Mock(method="POST", body=json.dumps({
                "display_name": "Test Block",
                "points": '-10',
                "weight": 11}).encode('utf-8')))
            assert block.points == orig_score

            # Test float doesn't work
            block.save_sga(mock.Mock(method="POST", body=json.dumps({
                "display_name": "Test Block",
                "points": '24.5',
                "weight": 11}).encode('utf-8')))
            assert block.points == orig_score

        orig_score = 23
        block = self.make_xblock()
        block.save_sga(mock.Mock(body='{}'))
        assert block.display_name == "Staff Graded Assignment"
        assert block.points == 100
        assert block.weight is None
        block.save_sga(mock.Mock(method="POST", body=json.dumps({
            "display_name": "Test Block",
            "points": orig_score,
            "weight": 11}).encode('utf-8')))
        assert block.display_name == "Test Block"
        assert block.points == orig_score
        assert block.weight == 11

        point_positive_int_test()
        weights_positive_float_test()

    @mock.patch('edx_sga.sga.StaffGradedAssignmentXBlock.get_student_item_dict')
    @mock.patch('edx_sga.sga.StaffGradedAssignmentXBlock.upload_allowed')
    @mock.patch('edx_sga.sga.get_sha1')
    def test_upload_download_assignment(self, get_sha1, upload_allowed, get_student_item_dict):
        # pylint: disable=unused-argument
        """
        Tests upload and download assignment for non staff.
        """
        file_name = 'test.txt'
        block = self.make_xblock()
        get_student_item_dict.return_value = {
            "student_id": 1,
            "course_id": block.block_course_id,
            "item_id": block.block_id,
            "item_type": 'sga',
        }
        upload_allowed.return_value = True

        with self.dummy_upload(file_name) as (upload, expected):
            with mock.patch('submissions.api.create_submission') as mocked_create_submission, mock.patch(
                "edx_sga.sga.StaffGradedAssignmentXBlock.file_storage_path",
                return_value=block.file_storage_path(SHA1, file_name)
            ), mock.patch(
                'edx_sga.sga.StaffGradedAssignmentXBlock.student_state', return_value={}
            ), mock.patch(
                'edx_sga.sga.StaffGradedAssignmentXBlock.get_or_create_student_module',
                return_value=fake_student_module()
            ) as mocked_create_student_module:
                block.upload_assignment(mock.Mock(params={'assignment': upload}))
            assert mocked_create_submission.called is True
            assert mocked_create_student_module.called is True

            with mock.patch(
                'edx_sga.sga.StaffGradedAssignmentXBlock.get_submission',
                return_value=fake_upload_submission(upload)
            ), mock.patch(
                "edx_sga.sga.StaffGradedAssignmentXBlock.file_storage_path",
                return_value=block.file_storage_path(SHA1, file_name)
            ):
                response = block.download_assignment(None)
                assert response.body == expected

            with mock.patch(
                "edx_sga.sga.StaffGradedAssignmentXBlock.file_storage_path",
                return_value=block.file_storage_path("", "test_notfound.txt")
            ), mock.patch(
                'edx_sga.sga.StaffGradedAssignmentXBlock.get_submission',
                return_value=fake_upload_submission(upload)
            ):
                response = block.download_assignment(None)
                assert response.status_code == 404

    @mock.patch('edx_sga.sga.StaffGradedAssignmentXBlock.get_student_item_dict')
    @mock.patch('edx_sga.sga.StaffGradedAssignmentXBlock.upload_allowed')
    @data(({'finalized': False}, True), ({}, True), ({'finalized': True}, False))
    @unpack
    def test_finalize_uploaded_assignment(
            self, finalized_setting, model_change_expected, upload_allowed, get_student_item_dict
    ):
        """
        Tests that finalize_uploaded_assignment sets a submission to be finalized
        """
        block = self.make_xblock()
        get_student_item_dict.return_value = {
            "student_id": 1,
            "course_id": block.block_course_id,
            "item_id": block.block_id,
            "item_type": 'sga',
        }
        upload_allowed.return_value = True
        existing_submitted_at_value = django_now()
        fake_submission_data = fake_get_submission(**finalized_setting)
        fake_submission_object = mock.Mock(
            submitted_at=existing_submitted_at_value,
            answer=fake_submission_data['answer']
        )

        with mock.patch(
            'edx_sga.sga.Submission.objects.get', return_value=fake_submission_object
        ), mock.patch(
            'edx_sga.sga.StaffGradedAssignmentXBlock.get_submission', return_value=fake_submission_data
        ), mock.patch(
            'edx_sga.sga.StaffGradedAssignmentXBlock.student_state', return_value={}
        ):
            block.finalize_uploaded_assignment(mock.Mock())

        assert fake_submission_object.answer['finalized'] is True
        assert (existing_submitted_at_value != fake_submission_object.submitted_at) is model_change_expected
        assert fake_submission_object.save.called is model_change_expected

    @mock.patch('edx_sga.sga.StaffGradedAssignmentXBlock.get_student_module')
    @mock.patch('edx_sga.sga.StaffGradedAssignmentXBlock.is_course_staff')
    @mock.patch('edx_sga.sga.get_sha1')
    def test_staff_upload_download_annotated(self, get_sha1, is_course_staff, get_student_module):
        # pylint: disable=no-member
        """
        Tests upload and download of annotated staff files.
        """
        get_student_module.return_value = fake_student_module()
        is_course_staff.return_value = True
        get_sha1.return_value = SHA1
        file_name = 'test.txt'
        block = self.make_xblock()

        with self.dummy_upload(file_name) as (upload, expected), mock.patch(
            "edx_sga.sga.StaffGradedAssignmentXBlock.staff_grading_data",
            return_value={}
        ) as staff_grading_data:
            block.staff_upload_annotated(mock.Mock(params={'annotated': upload, 'module_id': 1}))
        assert staff_grading_data.called is True

        with mock.patch(
            "edx_sga.sga.StaffGradedAssignmentXBlock.file_storage_path",
            return_value=block.file_storage_path(SHA1, file_name)
        ):
            response = block.staff_download_annotated(mock.Mock(params={'module_id': 1}))
            assert response.body == expected

        with mock.patch(
            "edx_sga.sga.StaffGradedAssignmentXBlock.file_storage_path",
            return_value=block.file_storage_path("", "test_notfound.txt")
        ):
            response = block.staff_download_annotated(
                mock.Mock(params={'module_id': 1})
            )
            assert response.status_code == 404

    @mock.patch('edx_sga.sga.StaffGradedAssignmentXBlock.get_student_module')
    @mock.patch('edx_sga.sga.StaffGradedAssignmentXBlock.is_course_staff')
    @mock.patch('edx_sga.sga.get_sha1')
    def test_download_annotated(self, get_sha1, is_course_staff, get_student_module):
        # pylint: disable=no-member
        """
        Test download annotated assignment for non staff.
        """
        get_student_module.return_value = fake_student_module()
        is_course_staff.return_value = True
        get_sha1.return_value = SHA1

        file_name = 'test.txt'
        block = self.make_xblock()

        with self.dummy_upload(file_name) as (upload, expected):
            with mock.patch(
                "edx_sga.sga.StaffGradedAssignmentXBlock.staff_grading_data",
                return_value={}
            ) as staff_grading_data:
                block.staff_upload_annotated(mock.Mock(params={
                    'annotated': upload,
                    'module_id': 1
                }))
            assert staff_grading_data.called is True
            self.personalize_upload(block, upload)
        with mock.patch(
            "edx_sga.sga.StaffGradedAssignmentXBlock.file_storage_path",
            return_value=block.file_storage_path(SHA1, file_name)
        ):
            response = block.download_annotated(None)
            assert response.body == expected

        with mock.patch(
            "edx_sga.sga.StaffGradedAssignmentXBlock.file_storage_path",
            return_value=block.file_storage_path("", "test_notfound.txt")
        ):
            response = block.download_annotated(None)
            assert response.status_code == 404

    @mock.patch('edx_sga.sga.StaffGradedAssignmentXBlock.upload_allowed')
    @mock.patch('edx_sga.sga.StaffGradedAssignmentXBlock.get_student_module')
    @mock.patch('edx_sga.sga.StaffGradedAssignmentXBlock.is_course_staff')
    @mock.patch('edx_sga.sga.get_sha1')
    def test_staff_download(self, get_sha1, is_course_staff, get_student_module, upload_allowed):
        """
        Test download for staff.
        """
        get_student_module.return_value = fake_student_module()
        is_course_staff.return_value = True
        upload_allowed.return_value = True
        get_sha1.return_value = SHA1
        block = self.make_xblock()

        with self.dummy_upload('test.txt') as (upload, expected), mock.patch(
            'edx_sga.sga.StaffGradedAssignmentXBlock.student_state', return_value={}
        ), mock.patch(
            'edx_sga.sga.StaffGradedAssignmentXBlock.get_or_create_student_module',
            return_value=fake_student_module()
        ), mock.patch(
            'submissions.api.create_submission'
        ) as mocked_create_submission:
            block.upload_assignment(mock.Mock(params={'assignment': upload}))
        assert mocked_create_submission.called is True
        self.personalize_upload(block, upload)

        with mock.patch(
            'edx_sga.sga.StaffGradedAssignmentXBlock.get_submission',
            return_value=fake_upload_submission(upload)
        ):
            response = block.staff_download(mock.Mock(params={
                'student_id': 1}))
            assert response.body == expected

        with mock.patch(
            "edx_sga.sga.StaffGradedAssignmentXBlock.file_storage_path",
            return_value=block.file_storage_path("", "test_notfound.txt")
        ), mock.patch(
            'edx_sga.sga.StaffGradedAssignmentXBlock.get_submission',
            return_value=fake_upload_submission(upload)
        ):
            response = block.staff_download(
                mock.Mock(params={'student_id': 1})
            )
            assert response.status_code == 404

    @unpack
    @data(
        {'past_due': False, 'score': None, 'is_finalized_submission': False, 'expected_value': True},
        {'past_due': True, 'score': None, 'is_finalized_submission': False, 'expected_value': False},
        {'past_due': False, 'score': 80, 'is_finalized_submission': False, 'expected_value': False},
        {'past_due': False, 'score': None, 'is_finalized_submission': True, 'expected_value': False},
    )
    def test_upload_allowed(self, past_due, score, is_finalized_submission, expected_value):
        """
        Tests that upload_allowed returns the right value under certain conditions
        """
        block = self.make_xblock()
        with mock.patch(
            "edx_sga.sga.StaffGradedAssignmentXBlock.past_due",
            return_value=past_due
        ), mock.patch(
            "edx_sga.sga.StaffGradedAssignmentXBlock.get_score",
            return_value=score
        ), mock.patch(
            "edx_sga.sga.is_finalized_submission",
            return_value=is_finalized_submission
        ):
            assert block.upload_allowed(submission_data={}) is expected_value

    @mock.patch('edx_sga.sga.StaffGradedAssignmentXBlock.count_archive_files')
    @mock.patch('edx_sga.sga.zip_student_submissions')
    @mock.patch('edx_sga.sga.StaffGradedAssignmentXBlock.get_sorted_submissions')
    @data((False, False), (True, True))
    @unpack
    def test_prepare_download_submissions(
            self,
            is_zip_file_available,
            downloadable,
            get_sorted_submissions,
            zip_student_submissions,
            count_archive_files
    ):
        """
        Test prepare download api
        """
        block = self.make_xblock()
        count_archive_files.return_value = 2
        get_sorted_submissions.return_value = [
            {
                'submission_id': uuid.uuid4().hex,
                'filename': "test_{}.txt".format(uuid.uuid4().hex),
                'timestamp': datetime.datetime.now(tz=pytz.utc)
            } for __ in range(2)
        ]
        zip_student_submissions.delay = mock.Mock()
        with mock.patch(
            "edx_sga.sga.StaffGradedAssignmentXBlock.is_zip_file_available",
            return_value=is_zip_file_available
        ), mock.patch(
            'edx_sga.sga.StaffGradedAssignmentXBlock.get_real_user',
            return_value=self.staff
        ), mock.patch(
            'edx_sga.utils.default_storage.modified_time',
            return_value=datetime.datetime.now()
        ):
            response = block.prepare_download_submissions(None)
            response_body = json.loads(response.body.decode('utf-8'))
            assert response_body["downloadable"] is downloadable

    @mock.patch('edx_sga.sga.get_file_modified_time_utc')
    @mock.patch('edx_sga.sga.StaffGradedAssignmentXBlock.count_archive_files')
    @mock.patch('edx_sga.sga.zip_student_submissions')
    @mock.patch('edx_sga.sga.StaffGradedAssignmentXBlock.get_sorted_submissions')
    @data((2, True, False), (1, False, True))
    @unpack
    def test_prepare_download_submissions_when_student_score_reset(
            self,
            count_archive_files,
            downloadable,
            zip_task_called,
            get_sorted_submissions,
            zip_student_submissions,
            count_archive_files_mock,
            get_file_modified_time_utc
    ):
        """
        Test prepare download api
        """
        now = datetime.datetime.now(tz=pytz.utc)
        block = self.make_xblock()
        count_archive_files_mock.return_value = count_archive_files
        get_sorted_submissions.return_value = [
            {
                'submission_id': uuid.uuid4().hex,
                'filename': "test_{}.txt".format(uuid.uuid4().hex),
                'timestamp': now
            } for __ in range(2)
        ]
        get_file_modified_time_utc.return_value = now
        zip_student_submissions.delay = mock.Mock()
        with mock.patch(
            "edx_sga.sga.StaffGradedAssignmentXBlock.is_zip_file_available", return_value=True
        ), mock.patch(
            'edx_sga.sga.StaffGradedAssignmentXBlock.get_real_user', return_value=self.staff
        ), mock.patch(
            'edx_sga.utils.default_storage.modified_time', return_value=datetime.datetime.now()
        ):
            response = block.prepare_download_submissions(None)
            response_body = json.loads(response.body.decode('utf-8'))
            assert response_body["downloadable"] is downloadable
            assert zip_student_submissions.delay.called is zip_task_called

    @mock.patch('edx_sga.sga.zip_student_submissions')
    @mock.patch('edx_sga.sga.StaffGradedAssignmentXBlock.get_sorted_submissions')
    def test_prepare_download_submissions_task_called(
            self,
            get_sorted_submissions,
            zip_student_submissions
    ):
        """
        Test prepare download api
        """
        block = self.make_xblock()
        get_sorted_submissions.return_value = [
            {
                'submission_id': uuid.uuid4().hex,
                'filename': "test_{}.txt".format(uuid.uuid4().hex),
                'timestamp': datetime.datetime.utcnow()
            } for __ in range(2)
        ]
        zip_student_submissions.delay = mock.Mock()
        with mock.patch(
            "edx_sga.sga.StaffGradedAssignmentXBlock.is_zip_file_available",
            return_value=False
        ), mock.patch(
            'edx_sga.sga.StaffGradedAssignmentXBlock.get_real_user',
            return_value=self.staff
        ), mock.patch(
            'edx_sga.sga.default_storage.modified_time',
            return_value=datetime.datetime.now()
        ):
            response = block.prepare_download_submissions(None)
            response_body = json.loads(response.body.decode('utf-8'))
            assert response_body["downloadable"] is False

        zip_student_submissions.delay.assert_called_once_with(
            six.text_type(block.block_course_id),
            six.text_type(block.block_id),
            six.text_type(block.location),
            self.staff.username
        )

    @data((False, False), (True, True))
    @unpack
    def test_download_submissions_status(self, is_zip_file_available, downloadable):
        """test download_submissions_status api"""
        block = self.make_xblock()
        with mock.patch(
            "edx_sga.sga.StaffGradedAssignmentXBlock.is_zip_file_available",
            return_value=is_zip_file_available
        ):
            response = block.download_submissions_status(None)
            response_body = json.loads(response.body.decode('utf-8'))
            assert response_body["zip_available"] is downloadable

    @mock.patch('edx_sga.sga.StaffGradedAssignmentXBlock.is_course_staff')
    def test_download_submissions(self, is_course_staff):
        """tests download_submissions"""
        block = self.make_xblock()
        is_course_staff.return_value = True

        expected = b"some information blah"
        filename = "foo.zip"
        path = os.path.join(self.temp_directory, filename)
        with open(path, "wb") as temp_file:
            temp_file.write(expected)

        with mock.patch(
            "edx_sga.sga.get_zip_file_path", return_value=path
        ), mock.patch(
            'edx_sga.sga.StaffGradedAssignmentXBlock.get_real_user',
            return_value=self.staff
        ), mock.patch(
            "edx_sga.sga.get_zip_file_name", return_value=filename
        ):
            response = block.download_submissions(None)
            assert response.status_code == 200
            assert response.body == expected

    def test_clear_student_state(self):
        """Tests that a student's state in the given problem is properly cleared"""
        block = self.make_xblock()
        orig_file_name = 'test.txt'
        fake_submission = fake_get_submission(filename=orig_file_name)
        uploaded_file_path = block.file_storage_path(SHA1, orig_file_name)

        with self.dummy_file_in_storage(uploaded_file_path) as file_path:
            with mock.patch(
                "edx_sga.sga.submissions_api.get_submissions",
                return_value=[fake_submission]
            ) as mocked_get_submissions, mock.patch(
                "edx_sga.sga.submissions_api.reset_score"
            ) as mocked_reset_score:
                assert self.default_storage.exists(file_path) is True
                block.clear_student_state(user_id=123)
                assert mocked_get_submissions.called is True
                # Clearing the student state should call 'reset_score' in the submission API,
                # which effectively resets the Submission record.
                assert mocked_reset_score.called is True
                # Clearing the student state should also delete the uploaded file
                assert self.default_storage.exists(file_path) is False



"""Shared test functionality"""
from __future__ import absolute_import

import hashlib
import shutil
import unittest
from contextlib import contextmanager
from datetime import datetime
from tempfile import mkdtemp

from lxml import etree
from mock import Mock

import pytz
from django.conf import settings
from django.core.files.base import ContentFile
from django.core.files.storage import default_storage
from xblock.fields import DateTime


class TempfileMixin(unittest.TestCase):
    """
    Test class that sets up a temp directory for 'uploaded' files, and configures Django's
    default_storage to use that directory
    """
    temp_directory = None
    default_storage = None
    _original_media_root = None
    _original_file_storage = None

    @classmethod
    def set_up_temp_directory(cls):
        """
        Creates a temp directory and fixes Django settings
        """
        cls._original_media_root = settings.MEDIA_ROOT
        cls._original_file_storage = settings.DEFAULT_FILE_STORAGE
        cls.temp_directory = mkdtemp()
        settings.MEDIA_ROOT = cls.temp_directory
        settings.DEFAULT_FILE_STORAGE = 'django.core.files.storage.FileSystemStorage'
        cls.default_storage = default_storage

    @classmethod
    def tear_down_temp_directory(cls):
        """
        Cleans up temp directory and fixes Django settings
        """
        shutil.rmtree(cls.temp_directory, ignore_errors=True)
        settings.MEDIA_ROOT = cls._original_media_root
        settings.DEFAULT_FILE_STORAGE = cls._original_file_storage
        del cls._original_media_root
        del cls._original_file_storage

    @contextmanager
    def dummy_upload(self, filename, data=b"some information"):
        """
        Provides a temporary file to act as a file that a user has uploaded

        Args:
            filename (str): A filename
            data (bytes): Random string

        Yields:
            (upload, data): The upload object and the data in the file
        """
        try:
            default_storage.save(filename, ContentFile(data))
            with default_storage.open(filename, "rb") as f:
                yield Mock(file=f), data
        finally:
            if default_storage.exists(filename):
                default_storage.delete(filename)

    @contextmanager
    def dummy_file_in_storage(self, rel_file_path):
        """
        Puts an empty file at the given file path (relative to the class temp directory)
        """
        try:
            default_storage.save(rel_file_path, ContentFile(b''))
            yield rel_file_path
        finally:
            if default_storage.exists(rel_file_path):
                default_storage.delete(rel_file_path)


class DummyResource(object):
    """
     A Resource class for use in tests
    """
    def __init__(self, path):
        self.path = path

    def __eq__(self, other):
        return isinstance(other, DummyResource) and self.path == other.path


def get_sha1(data):
    """
    Helper function to produce a SHA1 hash
    """
    algorithm = hashlib.sha1()
    algorithm.update(data)
    return algorithm.hexdigest()


def is_near_now(other_time):
    """
    Is a time pretty close to right now?
    """
    delta = abs(other_time - datetime.now(tz=pytz.utc))
    return delta.total_seconds() < 5


def parse_timestamp(timestamp):
    """
    Parse the xblock timestamp into a UTC datetime
    """
    return datetime.strptime(timestamp, DateTime.DATETIME_FORMAT).replace(
        tzinfo=pytz.utc
    )


def reformat_xml(xml_string):
    """
    Parse and render the XML to remove whitespace differences
    """
    parser = etree.XMLParser(remove_blank_text=True)
    root = etree.XML(xml_string, parser=parser)

    # Remove whitespace
    for elem in root.iter('*'):
        if elem.text is not None:
            elem.text = elem.text.strip()
            if not elem.text:
                elem.text = None

    return etree.tostring(root)



"""Constants"""

BLOCK_SIZE = 2**10 * 8  # 8kb
ITEM_TYPE = 'sga'


class ShowAnswer(object):
    """
    Constants for when to show answer
    """
    ALWAYS = "always"
    ANSWERED = "answered"
    ATTEMPTED = "attempted"
    CLOSED = "closed"
    FINISHED = "finished"
    CORRECT_OR_PAST_DUE = "correct_or_past_due"
    PAST_DUE = "past_due"
    NEVER = "never"



"""celery async tasks"""
from __future__ import absolute_import

import hashlib
import logging
import os
import tempfile
import zipfile

from django.core.files.storage import default_storage
from edx_sga.constants import ITEM_TYPE
from edx_sga.utils import get_file_storage_path
from lms import CELERY_APP  # pylint: disable=no-name-in-module
from opaque_keys.edx.locator import BlockUsageLocator
from student.models import user_by_anonymous_id
from submissions import api as submissions_api

log = logging.getLogger(__name__)


def _get_student_submissions(block_id, course_id, locator):
    """
    Returns valid submission file paths with the username of the student that submitted them.

    Args:
        course_id (unicode): edx course id
        block_id (unicode): edx block id
        locator (BlockUsageLocator): BlockUsageLocator for the sga module

    Returns:
        list(tuple): A list of 2-element tuples - (student username, submission file path)
    """
    submissions = submissions_api.get_all_submissions(
        course_id,
        block_id,
        ITEM_TYPE
    )
    return [
        (
            user_by_anonymous_id(submission['student_id']).username,
            get_file_storage_path(
                locator,
                submission['answer']['sha1'],
                submission['answer']['filename']
            )
        )
        for submission in submissions if submission['answer']
    ]


def _compress_student_submissions(zip_file_path, block_id, course_id, locator):
    """
    Creates a zip file of all student submissions for some course

    Args:
        destination_path (str): path (including name) of folder/file which we want to compress.
    """
    student_submissions = _get_student_submissions(block_id, course_id, locator)
    if not student_submissions:
        return

    log.info("Compressing %d student submissions to path: %s ", len(student_submissions), zip_file_path)
    # Build the zip file in memory using temporary file.
    with tempfile.TemporaryFile() as tmp:
        with zipfile.ZipFile(tmp, 'w', compression=zipfile.ZIP_DEFLATED) as zip_pointer:
            for student_username, submission_file_path in student_submissions:
                log.info(
                    "Creating zip file for student: %s, submission path: %s ",
                    student_username,
                    submission_file_path
                )
                with default_storage.open(submission_file_path, 'rb') as destination_file:
                    filename_in_zip = '{}_{}'.format(
                        student_username,
                        os.path.basename(submission_file_path)
                    )
                    zip_pointer.writestr(filename_in_zip, destination_file.read())
        # Reset file pointer
        tmp.seek(0)
        # Write the bytes of the in-memory zip file to an actual file
        log.info(
            "Moving zip file from memory to storage at path: %s ", zip_file_path
        )
        default_storage.save(zip_file_path, tmp)


@CELERY_APP.task
def zip_student_submissions(course_id, block_id, locator_unicode, username):
    """
    Task to download all submissions as zip file

    Args:
        course_id (unicode): edx course id
        block_id (unicode): edx block id
        locator_unicode (unicode): Unicode representing a BlockUsageLocator for the sga module
        username (unicode): user name of the staff user requesting the zip file
    """
    locator = BlockUsageLocator.from_string(locator_unicode)
    zip_file_path = get_zip_file_path(username, course_id, block_id, locator)
    log.info("Creating zip file for course: %s at path: %s", locator, zip_file_path)
    if default_storage.exists(zip_file_path):
        log.info("Deleting already-existing zip file at path: %s", zip_file_path)
        default_storage.delete(zip_file_path)
    _compress_student_submissions(
        zip_file_path,
        block_id,
        course_id,
        locator
    )


def get_zip_file_dir(locator):
    """
    Returns the relative directory path where we are saving the zipped submissions file.

    Args:
        locator (BlockUsageLocator): BlockUsageLocator for the sga module
    """
    return "{loc.org}/{loc.course}/{loc.block_type}_zipped".format(loc=locator)


def get_zip_file_name(username, course_id, block_id):
    """
    Returns the filename and extension of a submission zip file given a username and some
    information about the course.

    Args:
        username (unicode): staff user name
        course_id (unicode): edx course id
        block_id (unicode): edx block id
    """
    return "{username}_submissions_{id}_{course_key}.zip".format(
        username=username,
        id=hashlib.md5(block_id.encode('utf-8')).hexdigest(),
        course_key=course_id
    )


def get_zip_file_path(username, course_id, block_id, locator):
    """
    Returns the relative file path of a submission zip file given a username and some
    information about the course.

    Args:
        username (unicode): user name
        course_id (unicode): edx course id
        block_id (unicode): edx block id
        locator (BlockUsageLocator): BlockUsageLocator for the sga module
    """
    return os.path.join(
        get_zip_file_dir(locator),
        get_zip_file_name(username, course_id, block_id)
    )



"""XBlock mixins for showanswer"""
from __future__ import absolute_import

import logging

from django.utils.translation import ugettext_lazy as _
from edx_sga.constants import ShowAnswer
from xblock.fields import Scope, String

log = logging.getLogger(__name__)


class ShowAnswerXBlockMixin(object):
    """
    Mixin for showanswer functionality
    """

    showanswer = String(
        display_name=_("Show Answer"),
        help=_("Defines when to show the answer to the problem. "
               "A default value can be set in Advanced Settings."),
        scope=Scope.settings,
        default=ShowAnswer.PAST_DUE,  # Default to PAST_DUE instead of FINISHED for backwards compat
        values=[
            {"display_name": _("Always"), "value": ShowAnswer.ALWAYS},
            {"display_name": _("Answered"), "value": ShowAnswer.ANSWERED},
            {"display_name": _("Attempted"), "value": ShowAnswer.ATTEMPTED},
            {"display_name": _("Closed"), "value": ShowAnswer.CLOSED},
            {"display_name": _("Finished"), "value": ShowAnswer.FINISHED},
            {"display_name": _("Correct or Past Due"), "value": ShowAnswer.CORRECT_OR_PAST_DUE},
            {"display_name": _("Past Due"), "value": ShowAnswer.PAST_DUE},
            {"display_name": _("Never"), "value": ShowAnswer.NEVER}
        ]
    )
    solution = String(
        help=_("Solution to the problem to show to the user"),
        display_name=_("Solution"),
        scope=Scope.settings,
        multiline_editor='html',
        resettable_editor=False,
        default='',
    )

    def answer_available(self):  # pylint: disable=too-many-return-statements
        """
        Is the user allowed to see an answer?
        """
        if not self.correctness_available():
            # If correctness is being withheld, then don't show answers either.
            return False
        elif self.showanswer == '':
            return False
        elif self.showanswer == ShowAnswer.NEVER:
            return False
        elif self.runtime_user_is_staff():
            # This is after the 'never' check because admins can see the answer
            # unless the problem explicitly prevents it
            return True
        elif self.showanswer == ShowAnswer.ATTEMPTED:
            return self.has_attempted()
        elif self.showanswer == ShowAnswer.ANSWERED:
            return self.is_correct()
        elif self.showanswer == ShowAnswer.CLOSED:
            return self.closed()
        elif self.showanswer == ShowAnswer.FINISHED:
            return self.closed() or self.is_correct()
        elif self.showanswer == ShowAnswer.CORRECT_OR_PAST_DUE:
            return self.is_correct() or self.is_past_due()
        elif self.showanswer == ShowAnswer.PAST_DUE:
            return self.is_past_due()
        elif self.showanswer == ShowAnswer.ALWAYS:
            return True

        return False

    def closed(self):
        """
        Is the student still allowed to submit answers?
        """
        return not self.can_attempt() or self.is_past_due()

    def correctness_available(self):
        """
        Is the user allowed to see whether she's answered correctly?
        Limits access to the correct/incorrect flags, messages, and problem score.
        """
        raise NotImplementedError

    def is_past_due(self):
        """
        Is it now past this problem's due date, including grace period?
        """
        raise NotImplementedError

    def is_correct(self):
        """
        True iff full points
        """
        raise NotImplementedError

    def has_attempted(self):
        """
        True if the student has already attempted this problem
        """
        raise NotImplementedError

    def can_attempt(self):
        """
        True if the student is allowed to attempt the problem again
        """
        raise NotImplementedError

    def runtime_user_is_staff(self):
        """
        Is the current user a staff user?
        """
        raise NotImplementedError



"""Tests for ShowAnswerXBlockMixin"""
from __future__ import absolute_import

import datetime
from unittest import TestCase

import mock
import pytest

import ddt
import pytz
from edx_sga.constants import ShowAnswer
from edx_sga.showanswer import ShowAnswerXBlockMixin
from web_fragments.fragment import Fragment
from workbench.runtime import WorkbenchRuntime
from xblock.core import XBlock
from xblock.fields import DateTime, Integer, String

pytestmark = pytest.mark.django_db  # pylint: disable=invalid-name


class ShowAnswerXBlock(ShowAnswerXBlockMixin, XBlock):  # pylint: disable=abstract-method
    """
    A basic ShowAnswer XBlock implementation (for use in tests)
    """
    CATEGORY = 'showanswer'
    STUDIO_LABEL = 'Show Answers'

    color = String(default="red")
    count = Integer(default=42)
    comment = String(default="")
    date = DateTime(default=datetime.datetime(2014, 5, 14, tzinfo=pytz.UTC))
    editable_fields = ('color', 'count', 'comment', 'date')

    def student_view(self, context):  # pylint: disable=unused-argument
        """Just a placeholder"""
        return Fragment()


@ddt.ddt
class TestShowAnswerXBlock(TestCase):
    """
    Unit tests for ShowAnswerXBlockMixin
    """

    def setUp(self):
        super(TestShowAnswerXBlock, self).setUp()

        self.runtime = WorkbenchRuntime()

    def get_root(self):
        """Create an xblock block with a ShowAnswerXBlockMixin"""
        root_id = self.runtime.parse_xml_string('<showanswer />')
        return self.runtime.get_block(root_id)

    @ddt.data(*[
        [True, True, True],
        [True, False, False],
        [False, True, True],
        [False, False, True],
    ])
    @ddt.unpack
    @XBlock.register_temp_plugin(ShowAnswerXBlock, "showanswer")
    def test_closed(self, can_attempt, past_due, expected):
        """
        Assert possible values for closed()
        """
        block = self.get_root()
        with mock.patch.object(
            ShowAnswerXBlock, 'can_attempt', return_value=can_attempt,
        ), mock.patch.object(
            ShowAnswerXBlock, 'is_past_due', return_value=past_due,
        ):
            self.assertEqual(block.closed(), expected)

    @XBlock.register_temp_plugin(ShowAnswerXBlock, "showanswer")
    def test_answer_available_no_correctness(self):
        """
        If no correctness is available, the answer is not available
        """
        block = self.get_root()
        with mock.patch.object(
            ShowAnswerXBlock, 'correctness_available', return_value=False
        ):
            self.assertFalse(block.answer_available())

    @XBlock.register_temp_plugin(ShowAnswerXBlock, "showanswer")
    def test_answer_available_user_is_staff(self):
        """
        If user is staff and correctness is available, the answer is available
        """
        block = self.get_root()
        self.runtime.user_is_staff = True
        with mock.patch.object(
            ShowAnswerXBlock, 'correctness_available', return_value=True
        ), mock.patch.object(
            ShowAnswerXBlock, 'runtime_user_is_staff', return_value=True,
        ):
            self.assertTrue(block.answer_available())

    @ddt.data(*[
        ['', {}, False],
        [ShowAnswer.NEVER, {}, False],
        [ShowAnswer.ATTEMPTED, {}, False],
        [ShowAnswer.ATTEMPTED, {'has_attempted': True}, True],
        [ShowAnswer.ANSWERED, {}, False],
        [ShowAnswer.ANSWERED, {'is_correct': True}, True],
        [ShowAnswer.CLOSED, {}, False],
        [ShowAnswer.CLOSED, {'closed': True}, True],
        [ShowAnswer.FINISHED, {}, False],
        [ShowAnswer.FINISHED, {'closed': True}, True],
        [ShowAnswer.FINISHED, {'is_correct': True}, True],
        [ShowAnswer.FINISHED, {'closed': True, 'is_correct': True}, True],
        [ShowAnswer.CORRECT_OR_PAST_DUE, {}, False],
        [ShowAnswer.CORRECT_OR_PAST_DUE, {'is_correct': True}, True],
        [ShowAnswer.CORRECT_OR_PAST_DUE, {'is_past_due': True}, True],
        [ShowAnswer.CORRECT_OR_PAST_DUE, {'is_correct': True, 'is_past_due': True}, True],
        [ShowAnswer.PAST_DUE, {}, False],
        [ShowAnswer.PAST_DUE, {'is_past_due': True}, True],
        [ShowAnswer.ALWAYS, {}, True],
        ['unexpected', {}, False],
    ])
    @ddt.unpack
    @XBlock.register_temp_plugin(ShowAnswerXBlock, "showanswer")
    def test_answer_available_showanswer(self, showanswer, properties, expected):
        """
        Try different values of showanswer and assert answer_available()
        """
        block = self.get_root()
        block.showanswer = showanswer
        with mock.patch.object(
            ShowAnswerXBlock, 'correctness_available', return_value=True,
        ), mock.patch.object(
            ShowAnswerXBlock, 'runtime_user_is_staff', return_value=False,
        ), mock.patch.object(
            ShowAnswerXBlock, 'has_attempted', return_value=properties.get('has_attempted', False),
        ), mock.patch.object(
            ShowAnswerXBlock, 'is_correct', return_value=properties.get('is_correct', False),
        ), mock.patch.object(
            ShowAnswerXBlock, 'closed', return_value=properties.get('closed', False),
        ), mock.patch.object(
            ShowAnswerXBlock, 'is_past_due', return_value=properties.get('is_past_due', False),
        ):
            self.assertEqual(block.answer_available(), expected)



"""
Django command which migrates existing SGA submissions for a course from all
old SGA implementation before v0.4.0 to newer version that uses the
'submissions' application.
"""
from __future__ import absolute_import

import json

from django.core.management.base import BaseCommand, CommandError
from lms.djangoapps.courseware.courses import get_course_by_id
from lms.djangoapps.courseware.models import StudentModule
from opaque_keys.edx.keys import CourseKey
from student.models import anonymous_id_for_user
from submissions import api as submissions_api
from xmodule.modulestore.django import modulestore


class Command(BaseCommand):
    """
    Migrates existing SGA submissions for a course from old SGA implementation
    to newer version that uses the 'submissions' application.
    """
    args = "<course_id>"
    help = __doc__

    def handle(self, *args, **__options):
        """
        Migrates existing SGA submissions.
        """
        if not args:
            raise CommandError('Please specify the course id.')
        if len(args) > 1:
            raise CommandError('Too many arguments.')
        course_id = args[0]
        course_key = CourseKey.from_string(course_id)
        course = get_course_by_id(course_key)

        student_modules = StudentModule.objects.filter(
            course_id=course.id
        ).filter(
            module_state_key__contains='edx_sga'
        )

        blocks = {}
        for student_module in student_modules:
            block_id = student_module.module_state_key
            if block_id.block_type != 'edx_sga':
                continue
            block = blocks.get(block_id)
            if not block:
                blocks[block_id] = block = modulestore().get_item(block_id)
            state = json.loads(student_module.state)
            sha1 = state.get('uploaded_sha1')
            if not sha1:
                continue
            student = student_module.student
            submission_id = block.student_submission_id(
                anonymous_id_for_user(student, course.id))
            answer = {
                "sha1": sha1,
                "filename": state.get('uploaded_filename'),
                "mimetype": state.get('uploaded_mimetype'),
            }
            submission = submissions_api.create_submission(
                submission_id, answer)
            score = state.get('score')  # float
            if score:
                submissions_api.set_score(
                    submission['uuid'], int(score), block.max_score())



# -*- coding: utf-8 -*-
"""
Tests for SGA
"""
from __future__ import absolute_import

import datetime
import html
import json
import os
import shutil
import tempfile

import mock
import six.moves.urllib.error
import six.moves.urllib.parse
import six.moves.urllib.request

import pytz
from ddt import data, ddt, unpack
from django.contrib.auth.models import User
from django.core.exceptions import PermissionDenied
from django.db import transaction
from django.test.utils import override_settings
from edx_sga.constants import ShowAnswer
from edx_sga.sga import StaffGradedAssignmentXBlock
from edx_sga.tests.common import (DummyResource, TempfileMixin, get_sha1,
                                  is_near_now, parse_timestamp, reformat_xml)
from lms.djangoapps.courseware import module_render as render
from lms.djangoapps.courseware.models import StudentModule
from lms.djangoapps.courseware.tests.factories import StaffFactory
from opaque_keys.edx.locations import Location
from opaque_keys.edx.locator import CourseLocator
from student.models import UserProfile, anonymous_id_for_user
from student.tests.factories import AdminFactory
from submissions import api as submissions_api
from submissions.models import StudentItem
from xblock.field_data import DictFieldData
from xblock.fields import ScopeIds
from xmodule.modulestore.django import modulestore
from xmodule.modulestore.tests.django_utils import ModuleStoreTestCase
from xmodule.modulestore.tests.factories import CourseFactory, ItemFactory
from xmodule.modulestore.xml_exporter import export_course_to_xml
from xmodule.modulestore.xml_importer import import_course_from_xml


@ddt
class StaffGradedAssignmentXblockTests(TempfileMixin, ModuleStoreTestCase):
    """
    Create a SGA block with mock data.
    """
    @classmethod
    def setUpClass(cls):
        super(StaffGradedAssignmentXblockTests, cls).setUpClass()
        cls.set_up_temp_directory()

    @classmethod
    def tearDownClass(cls):
        super(StaffGradedAssignmentXblockTests, cls).tearDownClass()
        cls.tear_down_temp_directory()

    def setUp(self):
        """
        Creates a test course ID, mocks the runtime, and creates a fake storage
        engine for use in all tests
        """
        super(StaffGradedAssignmentXblockTests, self).setUp()
        self.course = CourseFactory.create(org='foo', number='bar', display_name='baz')
        self.descriptor = ItemFactory(category="pure", parent=self.course)
        self.course_id = self.course.id
        self.instructor = StaffFactory.create(course_key=self.course_id)
        self.student_data = mock.Mock()
        self.staff = AdminFactory.create(password="test")
        self.runtime = self.make_runtime()
        self.scope_ids = self.make_scope_ids(self.runtime)

    def make_runtime(self, **kwargs):
        """
        Make a runtime
        """
        runtime, _ = render.get_module_system_for_user(
            self.instructor,
            self.student_data,
            self.descriptor,
            self.course.id,
            mock.Mock(),
            mock.Mock(),
            mock.Mock(),
            course=self.course,
            # not sure why this isn't working, if set to true it looks for
            # 'display_name_with_default_escaped' field that doesn't exist in SGA
            wrap_xmodule_display=False,
            **kwargs
        )
        return runtime

    def make_scope_ids(self, runtime):
        """
        Make scope ids
        """
        # Not sure if this is a valid block type, might be sufficient for testing purposes
        block_type = 'sga'
        def_id = runtime.id_generator.create_definition(block_type)
        return ScopeIds(
            'user', block_type, def_id, self.descriptor.location
        )

    def make_one(self, display_name=None, **kw):
        """
        Creates a XBlock SGA for testing purpose.
        """
        field_data = DictFieldData(kw)
        block = StaffGradedAssignmentXBlock(self.runtime, field_data, self.scope_ids)
        block.location = Location(
            'foo', 'bar', 'baz', 'category', 'name', 'revision'
        )

        block.xmodule_runtime = self.runtime
        block.course_id = self.course_id
        block.category = 'problem'

        if display_name:
            block.display_name = display_name

        block.start = datetime.datetime(2010, 5, 12, 2, 42, tzinfo=pytz.utc)
        modulestore().create_item(
            self.staff.username, block.location.course_key, block.location.block_type, block.location.block_id
        )
        return block

    def make_student(self, block, name, make_state=True, **state):
        """
        Create a student along with submission state.
        """
        answer = {}
        module = None
        for key in ('sha1', 'mimetype', 'filename', 'finalized'):
            if key in state:
                answer[key] = state.pop(key)
        score = state.pop('score', None)

        with transaction.atomic():
            user = User(username=name, email='{}@example.com'.format(name))
            user.save()
            profile = UserProfile(user=user, name=name)
            profile.save()
            if make_state:
                module = StudentModule(
                    module_state_key=block.location,
                    student=user,
                    course_id=self.course_id,
                    state=json.dumps(state))
                module.save()

            anonymous_id = anonymous_id_for_user(user, self.course_id)
            item = StudentItem(
                student_id=anonymous_id,
                course_id=self.course_id,
                item_id=block.block_id,
                item_type='sga')
            item.save()

            if answer:
                student_id = block.get_student_item_dict(anonymous_id)
                submission = submissions_api.create_submission(student_id, answer)
                if score is not None:
                    submissions_api.set_score(
                        submission['uuid'], score, block.max_score())
            else:
                submission = None

            self.addCleanup(item.delete)
            self.addCleanup(profile.delete)
            self.addCleanup(user.delete)

            if make_state:
                self.addCleanup(module.delete)
                return {
                    'module': module,
                    'item': item,
                    'submission': submission
                }

            return {
                'item': item,
                'submission': submission
            }

    def personalize(self, block, module, item, submission):
        # pylint: disable=unused-argument
        """
        Set values on block from student state.
        """
        student_module = StudentModule.objects.get(pk=module.id)
        state = json.loads(student_module.state)
        for key, value in state.items():
            setattr(block, key, value)
        self.runtime.anonymous_student_id = item.student_id

    def test_ctor(self):
        """
        Test points are set correctly.
        """
        block = self.make_one(points=10)
        self.assertEqual(block.display_name, "Staff Graded Assignment")
        self.assertEqual(block.points, 10)

    def test_max_score(self):
        """
        Text max score is set correctly.
        """
        block = self.make_one(points=20)
        self.assertEqual(block.max_score(), 20)

    def test_max_score_integer(self):
        """
        Test assigning a float max score is rounded to nearest integer.
        """
        block = self.make_one(points=20.4)
        self.assertEqual(block.max_score(), 20)

    @override_settings(TECH_SUPPORT_EMAIL='foo@example.com')
    @mock.patch('edx_sga.sga._resource', DummyResource)
    @mock.patch('edx_sga.sga.render_template')
    @mock.patch('edx_sga.sga.Fragment')
    def test_student_view(self, fragment, render_template):
        # pylint: disable=unused-argument
        """
        Test student view renders correctly.
        """
        block = self.make_one("Custom name")
        self.personalize(block, **self.make_student(block, 'fred'))
        fragment = block.student_view()
        render_template.assert_called_once()
        template_arg = render_template.call_args[0][0]
        self.assertEqual(
            template_arg,
            'templates/staff_graded_assignment/show.html'
        )
        context = render_template.call_args[0][1]
        self.assertEqual(context['is_course_staff'], True)
        self.assertEqual(context['id'], 'name')
        self.assertEqual(context['support_email'], 'foo@example.com')
        student_state = json.loads(context['student_state'])
        self.assertEqual(
            student_state['display_name'],
            "Custom name"
        )
        self.assertEqual(student_state['uploaded'], None)
        self.assertEqual(student_state['annotated'], None)
        self.assertEqual(student_state['upload_allowed'], True)
        self.assertEqual(student_state['max_score'], 100)
        self.assertEqual(student_state['graded'], None)
        fragment.add_css.assert_called_once_with(
            DummyResource("static/css/edx_sga.css"))
        fragment.initialize_js.assert_called_once_with(
            "StaffGradedAssignmentXBlock")

    @mock.patch('edx_sga.sga._resource', DummyResource)
    @mock.patch('edx_sga.sga.render_template')
    @mock.patch('edx_sga.sga.Fragment')
    def test_student_view_with_upload(self, fragment, render_template):
        # pylint: disable=unused-argument
        """
        Test student is able to upload assignment correctly.
        """
        block = self.make_one()
        self.personalize(block, **self.make_student(
            block, 'fred"', sha1='foo', filename='foo.bar'))
        block.student_view()
        context = render_template.call_args[0][1]
        student_state = json.loads(context['student_state'])
        self.assertEqual(student_state['uploaded'], {'filename': 'foo.bar'})

    @mock.patch('edx_sga.sga._resource', DummyResource)
    @mock.patch('edx_sga.sga.render_template')
    @mock.patch('edx_sga.sga.Fragment')
    def test_student_view_with_annotated(self, fragment, render_template):
        # pylint: disable=unused-argument
        """
        Test student view shows annotated files correctly.
        """
        block = self.make_one(
            annotated_sha1='foo', annotated_filename='foo.bar')
        self.personalize(block, **self.make_student(block, "fred"))
        block.student_view()
        context = render_template.call_args[0][1]
        student_state = json.loads(context['student_state'])
        self.assertEqual(student_state['annotated'], {'filename': 'foo.bar'})

    @mock.patch('edx_sga.sga._resource', DummyResource)
    @mock.patch('edx_sga.sga.render_template')
    @mock.patch('edx_sga.sga.Fragment')
    def test_student_view_with_score(self, fragment, render_template):
        # pylint: disable=unused-argument
        """
        Tests scores are displayed correctly on student view.
        """
        block = self.make_one()
        self.personalize(block, **self.make_student(
            block, 'fred', filename='foo.txt', score=10))
        fragment = block.student_view()
        render_template.assert_called_once()
        template_arg = render_template.call_args[0][0]
        self.assertEqual(
            template_arg,
            'templates/staff_graded_assignment/show.html'
        )
        context = render_template.call_args[0][1]
        self.assertEqual(context['is_course_staff'], True)
        self.assertEqual(context['id'], 'name')
        student_state = json.loads(context['student_state'])
        self.assertEqual(
            student_state['display_name'],
            "Staff Graded Assignment"
        )
        self.assertEqual(student_state['uploaded'], {u'filename': u'foo.txt'})
        self.assertEqual(student_state['annotated'], None)
        self.assertEqual(student_state['upload_allowed'], False)
        self.assertEqual(student_state['max_score'], 100)
        self.assertEqual(student_state['graded'],
                         {u'comment': '', u'score': 10})
        fragment.add_css.assert_called_once_with(
            DummyResource("static/css/edx_sga.css"))
        fragment.initialize_js.assert_called_once_with(
            "StaffGradedAssignmentXBlock")

    def test_studio_view(self):
        # pylint: disable=unused-argument
        """
        Test studio view is using the StudioEditableXBlockMixin function
        """
        with mock.patch('edx_sga.sga.StudioEditableXBlockMixin.studio_view') as studio_view_mock:
            block = self.make_one()
            block.studio_view()
        studio_view_mock.assert_called_once_with(None)

    def test_save_sga(self):
        """
        Tests save SGA  block on studio.
        """
        def weights_positive_float_test():
            """
            tests weight is non negative float.
            """
            orig_weight = 11.0

            # Test negative weight doesn't work
            block.save_sga(mock.Mock(method="POST", body=json.dumps({
                "display_name": "Test Block",
                "points": '100',
                "weight": -10.0})))
            self.assertEqual(block.weight, orig_weight)

            # Test string weight doesn't work
            block.save_sga(mock.Mock(method="POST", body=json.dumps({
                "display_name": "Test Block",
                "points": '100',
                "weight": "a"})))
            self.assertEqual(block.weight, orig_weight)

        def point_positive_int_test():
            """
            Tests point is positive number.
            """
            # Test negative doesn't work
            block.save_sga(mock.Mock(method="POST", body=json.dumps({
                "display_name": "Test Block",
                "points": '-10',
                "weight": 11})))
            self.assertEqual(block.points, orig_score)

            # Test float doesn't work
            block.save_sga(mock.Mock(method="POST", body=json.dumps({
                "display_name": "Test Block",
                "points": '24.5',
                "weight": 11})))
            self.assertEqual(block.points, orig_score)

        orig_score = 23
        block = self.make_one()
        block.save_sga(mock.Mock(body='{}'))
        self.assertEqual(block.display_name, "Staff Graded Assignment")
        self.assertEqual(block.points, 100)
        self.assertEqual(block.weight, None)
        block.save_sga(mock.Mock(method="POST", body=json.dumps({
            "display_name": "Test Block",
            "points": str(orig_score),
            "weight": 11})))
        self.assertEqual(block.display_name, "Test Block")
        self.assertEqual(block.points, orig_score)
        self.assertEqual(block.weight, 11)

        point_positive_int_test()
        weights_positive_float_test()

    def test_upload_download_assignment(self):
        """
        Tests upload and download assignment for non staff.
        """
        block = self.make_one()
        self.personalize(block, **self.make_student(block, "fred"))
        with self.dummy_upload("text.txt") as (upload, expected):
            block.upload_assignment(mock.Mock(params={'assignment': upload}))
        response = block.download_assignment(None)
        self.assertEqual(response.body, expected)

        with mock.patch(
            "edx_sga.sga.StaffGradedAssignmentXBlock.file_storage_path",
            return_value=block.file_storage_path(  # lint-amnesty, pylint: disable=protected-access
                "", "test_notfound.txt"
            )  # lint-amnesty, pylint: disable=protected-access
        ):
            response = block.download_assignment(None)
            self.assertEqual(response.status_code, 404)

    def test_finalize_uploaded_assignment(self):
        """
        Tests that finalize_uploaded_assignment sets a submission to be finalized
        """
        block = self.make_one()
        created_student_data = self.make_student(block, "fred1", finalized=False, filename='answer')
        self.personalize(block, **created_student_data)
        submission_data = created_student_data['submission']
        response = block.finalize_uploaded_assignment(mock.Mock(method="POST"))
        recent_submission_data = block.get_submission()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json, block.student_state())
        self.assertEqual(submission_data['uuid'], recent_submission_data['uuid'])
        self.assertTrue(recent_submission_data['answer']['finalized'])

    def test_staff_upload_download_annotated(self):
        # pylint: disable=no-member
        """
        Tests upload and download of annotated staff files.
        """
        block = self.make_one()
        fred = self.make_student(block, "fred1")['module']
        with self.dummy_upload('test.txt') as (upload, expected):
            block.staff_upload_annotated(mock.Mock(params={
                'annotated': upload,
                'module_id': fred.id}))
        response = block.staff_download_annotated(mock.Mock(params={
            'module_id': fred.id}))
        self.assertEqual(response.body, expected)

        with mock.patch(
            "edx_sga.sga.StaffGradedAssignmentXBlock.file_storage_path",
            return_value=block.file_storage_path(  # lint-amnesty, pylint: disable=protected-access
                "", "test_notfound.txt"
            )  # lint-amnesty, pylint: disable=protected-access
        ):
            response = block.staff_download_annotated(
                mock.Mock(
                    params={
                        'module_id': fred.id
                    }
                )
            )
            self.assertEqual(response.status_code, 404)

    def test_staff_upload_annotated_state(self):
        # pylint: disable=no-member
        """
        Test state recorded in the module state when staff_upload_annotated is called
        """
        block = self.make_one()
        fred = self.make_student(block, "fred1")['module']

        with self.dummy_upload('testa.txt') as (upload, _):
            block.upload_assignment(mock.Mock(params={"assignment": upload}))
        block.finalize_uploaded_assignment(mock.Mock(method="POST"))

        with self.dummy_upload('testb.txt') as (upload, expected):
            request = mock.Mock(params={
                'annotated': upload,
                'module_id': fred.id
            })
            resp = block.staff_upload_annotated(request)
        assert resp.json == block.staff_grading_data()
        state = json.loads(block.get_student_module(fred.id).state)
        assert state['annotated_mimetype'] == 'text/plain'
        parsed_date = parse_timestamp(state['annotated_timestamp'])
        assert is_near_now(parsed_date)
        assert state['annotated_filename'].endswith('testb.txt')
        assert state['annotated_sha1'] == get_sha1(expected)

    def test_download_annotated(self):
        # pylint: disable=no-member
        """
        Test download annotated assignment for non staff.
        """
        block = self.make_one()
        students_info = [
            ['fred2', 'test_fred.txt', 'student 1'],
            ['foo2', 'test_foo.txt', 'student 2']
        ]
        students = []
        for student_name, filename, text in students_info:
            student = self.make_student(block, student_name)
            with self.dummy_upload(filename, text) as (upload, __):
                block.staff_upload_annotated(mock.Mock(params={
                    'annotated': upload,
                    'module_id': student['module'].id}))
            students.append((student, text))

        for student, text in students:
            self.personalize(block, **student)
            response = block.download_annotated(None)
            self.assertEqual(response.body, text)

        with mock.patch(
            "edx_sga.sga.StaffGradedAssignmentXBlock.file_storage_path",
            return_value=block.file_storage_path(  # lint-amnesty, pylint: disable=protected-access
                "",
                "test_notfound.txt"
            )  # lint-amnesty, pylint: disable=protected-access
        ):
            response = block.download_annotated(None)
            self.assertEqual(response.status_code, 404)

    def test_staff_download(self):
        """
        Test validate each student assignment for staff download.
        """
        students = []
        block = self.make_one()
        students_info = [
            ['fred2', 'test_fred.txt', 'student 1'],
            ['foo2', 'test_foo.txt', 'student 2']
        ]
        for student_name, filename, text in students_info:
            student = self.make_student(block, student_name)
            self.personalize(block, **student)
            with self.dummy_upload(filename, text) as (upload, __):
                block.upload_assignment(mock.Mock(params={'assignment': upload}))
            students.append((student, text,))

        for student, text in students:
            response = block.staff_download(
                mock.Mock(params={'student_id': student['item'].student_id})
            )
            self.assertEqual(response.body, text)

        # assert that staff cannot access invalid files
        for student, __ in students:
            with mock.patch(
                "edx_sga.sga.StaffGradedAssignmentXBlock.file_storage_path",
                return_value=block.file_storage_path(  # lint-amnesty, pylint: disable=protected-access
                    "",
                    "test_notfound.txt"
                )  # lint-amnesty, pylint: disable=protected-access
            ):
                response = block.staff_download(
                    mock.Mock(
                        params={
                            'student_id': student['item'].student_id
                        }
                    )
                )
                self.assertEqual(response.status_code, 404)

    def test_download_annotated_unicode_filename(self):
        """
        Tests download annotated assignment
        with filename in unicode for non staff member.
        """
        block = self.make_one()
        fred = self.make_student(block, "fred2")
        with self.dummy_upload('файл.txt') as (upload, expected):
            block.staff_upload_annotated(mock.Mock(params={
                'annotated': upload,
                'module_id': fred['module'].id}))
        self.personalize(block, **fred)
        response = block.download_annotated(None)
        self.assertEqual(response.body, expected)

        with mock.patch(
            "edx_sga.sga.StaffGradedAssignmentXBlock.file_storage_path",
            return_value=block.file_storage_path(  # lint-amnesty, pylint: disable=protected-access
                "",
                "test_notfound.txt"
            )  # lint-amnesty, pylint: disable=protected-access
        ):
            response = block.download_annotated(None)
            self.assertEqual(response.status_code, 404)

    def test_staff_download_unicode_filename(self):
        """
        Tests download assignment with filename in unicode for staff.
        """
        block = self.make_one()
        student = self.make_student(block, 'fred')
        self.personalize(block, **student)
        with self.dummy_upload('файл.txt') as (upload, expected):
            block.upload_assignment(mock.Mock(params={'assignment': upload}))
        response = block.staff_download(mock.Mock(params={
            'student_id': student['item'].student_id}))
        self.assertEqual(response.body, expected)

        with mock.patch(
            "edx_sga.sga.StaffGradedAssignmentXBlock.file_storage_path",
            return_value=block.file_storage_path(  # lint-amnesty, pylint: disable=protected-access
                "",
                "test_notfound.txt"
            )  # lint-amnesty, pylint: disable=protected-access
        ):
            response = block.staff_download(
                mock.Mock(
                    params={
                        'student_id': student['item'].student_id
                    }
                )
            )
            self.assertEqual(response.status_code, 404)

    def test_staff_download_filename_with_spaces(self):
        """
        Tests download assignment with spaces in filename.
        """
        file_name = 'my assignment 2.txt'
        block = self.make_one()
        student = self.make_student(block, 'fred')
        self.personalize(block, **student)
        with self.dummy_upload(file_name) as (upload, expected):
            block.upload_assignment(mock.Mock(params={'assignment': upload}))
        response = block.staff_download(mock.Mock(params={
            'student_id': student['item'].student_id}))
        self.assertEqual(response.body, expected)
        assert six.moves.urllib.parse.quote(file_name.encode('utf-8')) in response.content_disposition

    @data("my,assignment.txt", "my,1,1,assignment.txt")
    def test_file_download_comma_in_name(self, file_name):
        """
        Tests download assignment with comma in filename.
        """
        block = self.make_one()
        student = self.make_student(block, 'fred')
        self.personalize(block, **student)
        with self.dummy_upload(file_name) as (upload, expected):
            block.upload_assignment(mock.Mock(params={'assignment': upload}))
        response = block.staff_download(mock.Mock(params={
            'student_id': student['item'].student_id}))
        self.assertEqual(response.body, expected)
        assert six.moves.urllib.parse.quote(file_name.encode('utf-8')) in response.content_disposition

    def test_get_staff_grading_data_not_staff(self):
        """
        test staff grading data for non staff members.
        """
        self.runtime.user_is_staff = False
        block = self.make_one()
        with self.assertRaises(PermissionDenied):
            block.get_staff_grading_data(None)

    def test_get_staff_grading_data(self):
        # pylint: disable=no-member
        """
        Test fetch grading data for staff members.
        """
        block = self.make_one()
        barney = self.make_student(
            block, "barney",
            filename="foo.txt",
            score=10,
            annotated_filename="foo_corrected.txt",
            comment="Good work!")
        fred = self.make_student(
            block, "fred",
            filename="bar.txt")
        data = block.get_staff_grading_data(None).json_body  # lint-amnesty, pylint: disable=redefined-outer-name
        assignments = sorted(data['assignments'], key=lambda x: x['username'])

        barney_assignment, fred_assignment = assignments

        assert barney_assignment['module_id'] == barney['module'].id
        assert barney_assignment['username'] == 'barney'
        assert barney_assignment['fullname'] == 'barney'
        assert barney_assignment['filename'] == 'foo.txt'
        assert barney_assignment['score'] == 10
        assert barney_assignment['annotated'] == 'foo_corrected.txt'
        assert barney_assignment['comment'] == 'Good work!'
        assert barney_assignment['approved'] is True
        assert barney_assignment['finalized'] is True
        assert barney_assignment['may_grade'] is False
        assert barney_assignment['needs_approval'] is False
        assert barney_assignment['student_id'] == barney['item'].student_id
        assert barney_assignment['submission_id'] == barney['submission']['uuid']
        assert is_near_now(parse_timestamp(barney_assignment['timestamp']))

        assert fred_assignment['module_id'] == fred['module'].id
        assert fred_assignment['username'] == 'fred'
        assert fred_assignment['fullname'] == 'fred'
        assert fred_assignment['filename'] == 'bar.txt'
        assert fred_assignment['score'] is None
        assert fred_assignment['annotated'] == u''
        assert fred_assignment['comment'] == u''
        assert fred_assignment['approved'] is False
        assert fred_assignment['finalized'] is True
        assert fred_assignment['may_grade'] is True
        assert fred_assignment['needs_approval'] is False
        assert fred_assignment['student_id'] == fred['item'].student_id
        assert fred_assignment['submission_id'] == fred['submission']['uuid']
        assert is_near_now(parse_timestamp(fred_assignment['timestamp']))

    @mock.patch('edx_sga.sga.log')
    def test_assert_logging_when_student_module_created(self, mocked_log):
        """
        Verify logs are created when student modules are created.
        """
        block = self.make_one()
        self.make_student(
            block,
            "tester",
            make_state=False,
            filename="foo.txt",
            score=10,
            annotated_filename="foo_corrected.txt",
            comment="Good work!"
        )
        block.staff_grading_data()
        module_creation_log_message = mocked_log.info.call_args[0]
        self.assertIn('tester', module_creation_log_message)
        self.assertIn(block.course_id, module_creation_log_message)
        self.assertIn(block.location, module_creation_log_message)

    def test_enter_grade_instructor(self):
        # pylint: disable=no-member
        """
        Test enter grade by instructors.
        """
        block = self.make_one()
        block.is_instructor = lambda: True
        fred = self.make_student(block, "fred5", filename='foo.txt')
        block.enter_grade(mock.Mock(params={
            'module_id': fred['module'].id,
            'submission_id': fred['submission']['uuid'],
            'grade': 9,
            'comment': "Good!"}))
        state = json.loads(StudentModule.objects.get(
            pk=fred['module'].id).state)
        self.assertEqual(state['comment'], 'Good!')
        self.assertEqual(block.get_score(fred['item'].student_id), 9)

    def test_enter_grade_staff(self):
        # pylint: disable=no-member
        """
        Test grade enter by staff.
        """
        block = self.make_one()
        fred = self.make_student(block, "fred5", filename='foo.txt')
        block.enter_grade(mock.Mock(params={
            'module_id': fred['module'].id,
            'submission_id': fred['submission']['uuid'],
            'grade': 9,
            'comment': "Good!"}))
        state = json.loads(StudentModule.objects.get(
            pk=fred['module'].id).state)
        self.assertEqual(state['comment'], 'Good!')
        self.assertEqual(state['staff_score'], 9)

    @data(None, "", '9.24', "second")
    def test_enter_grade_fail(self, grade):
        # pylint: disable=no-member
        """
        Tests grade enter fail.
        """
        block = self.make_one()
        fred = self.make_student(block, "fred5", filename='foo.txt')
        with mock.patch('edx_sga.sga.log') as mocked_log:
            block.enter_grade(
                mock.Mock(
                    params={
                        'module_id': fred['module'].id,
                        'submission_id': fred['submission']['uuid'],
                        'grade': grade
                    }
                )
            )
        mocked_log.error.assert_called_with(
            "enter_grade: invalid grade submitted for course:%s module:%s student:%s",
            block.course_id,
            block.location,
            "fred5"
        )

    def test_remove_grade(self):
        # pylint: disable=no-member
        """
        Test remove grade.
        """
        block = self.make_one()
        student = self.make_student(
            block, "fred6", score=9, comment='Good!')
        module = student['module']
        item = student['item']
        request = mock.Mock(params={
            'module_id': module.id,
            'student_id': item.student_id,
        })
        block.remove_grade(request)
        state = json.loads(StudentModule.objects.get(pk=module.id).state)
        self.assertEqual(block.get_score(item.student_id), None)
        self.assertEqual(state['comment'], '')

    @data(True, False)
    def test_past_due(self, is_past):
        """
        Test due date is pass.
        """
        block = self.make_one()
        now = datetime.datetime.now(tz=pytz.utc)
        delta = datetime.timedelta(days=-1 if is_past else 1)
        block.due = now + delta
        assert block.past_due() is is_past
        assert block.past_due() is block.is_past_due()

    @data(True, False)
    def test_showanswer(self, is_answer_available):
        """
        The student state should have the solution if answer is available
        """
        block = self.make_one(solution='A solution')
        with mock.patch(
            'edx_sga.sga.StaffGradedAssignmentXBlock.answer_available',
            return_value=is_answer_available,
        ):
            assert block.student_state()['solution'] == ('A solution' if is_answer_available else '')

    @data(
        (True, '/static/foo'),
        (False, '/c4x/foo/bar/asset')
    )
    @unpack
    def test_replace_url(self, has_static_asset_path, path):
        """
        If the static asset path is set on a course, it should be substituted when the course is rendered
        """
        # make a runtime with a static asset path, which will override the base_asset_url
        if has_static_asset_path:
            self.runtime = self.make_runtime(static_asset_path='foo')
            self.scope_ids = self.make_scope_ids(self.runtime)

        block = self.make_one(
            solution='<a href="/static/test.pdf">A PDF</a>',
            showanswer=ShowAnswer.ALWAYS,
        )
        solution = block.student_state()['solution']
        assert '<a href="{}/test.pdf">A PDF</a>'.format(path) == solution

    def test_base_asset_url(self):
        """
        The base asset url for the course should be passed to the javascript so it can replace static links
        """
        block = self.make_one(solution='A solution')
        assert block.student_state()['base_asset_url'] == '/c4x/foo/bar/asset/'

    def test_correctness_available(self):
        """
        Correctness should always be available
        """
        block = self.make_one()
        assert block.correctness_available() is True

    def test_has_attempted(self):
        """
        A SGA problem is attempted if they uploaded their submission and it is finalized
        """
        block = self.make_one()
        assert block.has_attempted() is False
        assert block.is_correct() is False
        assert block.can_attempt() is True

        with self.dummy_upload('test.txt') as (upload, _):
            block.upload_assignment(mock.Mock(params={'assignment': upload}))
        assert block.has_attempted() is False
        assert block.is_correct() is False
        assert block.can_attempt() is True

        block.finalize_uploaded_assignment(mock.Mock())
        assert block.has_attempted() is True
        assert block.is_correct() is True
        assert block.can_attempt() is False

    @data(True, False)
    def test_runtime_user_is_staff(self, is_staff):
        course = CourseFactory.create(org='org', number='bar', display_name='baz')
        descriptor = ItemFactory(category="pure", parent=course)

        staff = StaffFactory.create(course_key=course.id)
        self.runtime, _ = render.get_module_system_for_user(
            staff if is_staff else User.objects.create(),
            self.student_data,
            descriptor,
            course.id,
            mock.Mock(),
            mock.Mock(),
            mock.Mock(),
            course=course
        )
        block = self.make_one()
        assert block.runtime_user_is_staff() is is_staff

    @data(True, False)
    def test_grace_period(self, has_grace_period):
        block = self.make_one()

        # Due date is slightly in the past
        block.due = datetime.datetime.now(tz=pytz.utc) - datetime.timedelta(seconds=500)
        if has_grace_period:
            block.graceperiod = datetime.timedelta(days=1)

        assert block.past_due() is not has_grace_period

    def make_test_vertical(self, solution_attribute=None, solution_element=None):
        """Create a test vertical with an SGA unit inside"""
        solution_attribute = 'solution="{}"'.format(html.escape(solution_attribute)) if solution_attribute else ''
        solution_element = '<solution>{}</solution>'.format(solution_element) if solution_element else ''

        return (
            """<vertical display_name="SGA Unit">
              <edx_sga url_name="edx_sga" xblock-family="xblock.v1" display_name="SGA Test 1" {solution_attribute}>
                {solution_element}
              </edx_sga>
            </vertical>""".format(solution_attribute=solution_attribute, solution_element=solution_element)
        )

    def import_test_course(self, solution_attribute=None, solution_element=None):
        """
        Import the test course with the sga unit
        """
        # adapted from edx-platform/cms/djangoapps/contentstore/management/commands/tests/test_cleanup_assets.py
        root = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
        input_dir = os.path.join(root, "test_data")

        temp_dir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(temp_dir))

        xml_dir = os.path.join(temp_dir, "xml")
        shutil.copytree(input_dir, xml_dir)

        with open(os.path.join(xml_dir, "2017_SGA", "vertical", "vertical.xml"), "w") as f:
            f.write(self.make_test_vertical(solution_attribute, solution_element))

        store = modulestore()
        import_course_from_xml(
            store,
            'sga_user',
            xml_dir,
        )

        return store.get_course(CourseLocator.from_string('SGAU/SGA101/course'))

    @data(*[
        ['<p>Broken xml', "<p>You're seeing the answer</p>", '<p>Broken xml'],
        ['<p>Broken xml', None, '<p>Broken xml'],
        [None, "<p>You're seeing the answer</p>", "<p>You're seeing the answer</p>"],
        [None, None, ''],
    ])
    @unpack
    def test_import(self, solution_attribute_value, solution_element_value, expected_solution_text):
        """Import the test course with the SGA module"""
        course = self.import_test_course(solution_attribute_value, solution_element_value)
        sga = course.get_children()[0].get_children()[0].get_children()[0].get_children()[0]
        assert expected_solution_text in sga.solution
        assert sga.showanswer == ShowAnswer.PAST_DUE

    @data(*[
        ['<p>Broken xml', "<p>You're seeing the answer</p>", '<p>Broken xml', None],
        ['<p>Broken xml', None, '<p>Broken xml', None],
        [None, "<p>You're seeing the answer</p>", None, "<p>You're seeing the answer</p>"],
        [None, None, None, None],
    ])
    @unpack
    def test_export(
            self, solution_attribute_value, solution_element_value,
            expected_solution_attribute, expected_solution_element):
        """Export the test course with the SGA module"""
        course = self.import_test_course(solution_attribute_value, solution_element_value)

        temp_dir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(temp_dir))

        store = modulestore()
        export_course_to_xml(store, None, course.id, temp_dir, "2017_SGA")

        with open(os.path.join(temp_dir, "2017_SGA", "vertical", "vertical.xml")) as f:
            content = f.read()

        # If both are true the expected output should only have the attribute, since it took precedence
        # and the attribute contents are broken XML
        assert reformat_xml(content) == reformat_xml(
            self.make_test_vertical(expected_solution_attribute, expected_solution_element)
        )



"""Setup for edx-sga XBlock."""

from __future__ import absolute_import

import os

from setuptools import find_packages, setup

import edx_sga


def package_data(pkg, root_list):
    """Generic function to find package_data for `pkg` under `root`."""
    data = []
    for root in root_list:
        for dirname, _, files in os.walk(os.path.join(pkg, root)):
            for fname in files:
                data.append(os.path.relpath(os.path.join(dirname, fname), pkg))

    return {pkg: data}

setup(
    name='edx-sga',
    version=edx_sga.__version__,
    description='edx-sga Staff Graded Assignment XBlock',
    license='Affero GNU General Public License v3 (GPLv3)',
    url="https://github.com/mitodl/edx-sga",
    author="MITx",
    zip_safe=False,
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'XBlock',
        'xblock-utils',
    ],
    entry_points={
        'xblock.v1': [
            'edx_sga = edx_sga.sga:StaffGradedAssignmentXBlock',
        ]
    },
    package_data=package_data("edx_sga", ["static", "templates"]),
)
